import pytest

from src.py_cdll import CDLL, ValueNotFoundError, MultipleValuesFoundError


def test_replace_unique_present_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data_new: str = "data_new"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)

    # Execution
    list0.replace_unique(value=data1, replacement=data_new)

    # Validation
    assert list0[1] is data_new


def test_replace_unique_not_present_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data_find: str = "data_find"
    data_new: str = "data_new"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)

    # Validation
    with pytest.raises(ValueNotFoundError):
        list0.replace_unique(value=data_find, replacement=data_new)


def test_replace_unique_multiple_options_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data_new: str = "data_new"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data1)
    list0.append(data2)

    # Validation
    with pytest.raises(MultipleValuesFoundError):
        list0.replace_unique(value=data1, replacement=data_new)



from typing import TypeVar, Iterator, Sequence, Callable

from .exceptions import NotANodeError, NodesNotInSameListError, NoNewHeadError

_T = TypeVar("_T")


class Node:
    """Circular Doubly Linked Node"""

    def __init__(self, value: _T) -> None:
        self.head: Node = self
        self.next: Node = self
        self.previous: Node = self
        self.index: int = 0
        self.value: _T = value

    def __repr__(self) -> str:
        values: list = []

        # start_index: int = self.index
        # current: Node = self
        # while True:
        #     values.append(current.value)
        #     current = current.next
        #     # if not isinstance(current, Node) or current is self:
        #     if current is self or current.index == start_index:
        #         break

        # string = f"Node(...{id(self) % 1000})[{self.head.previous.index + 1}, {[self.value]}{values[1:]}]"
        # string = f"Node(...{id(self) % 1000})[LEN, {[self.value]}{values[1:]}]"
        string = f"Node(...{id(self) % 1000})[{self.head.previous.index + 1}, VAL:VALS]"
        return string


def nodes_from_values(values: Sequence[_T]) -> Node | None:
    """Create circular doubly linked list of Nodes containing values."""
    values_iterator: Iterator[_T] = iter(values)

    try:
        head: Node | None = Node(value=next(values_iterator))
        current: Node | None = head
    except StopIteration:
        head, current = None, None

    for value in values_iterator:
        node: Node = Node(value=value)
        insert_between(before=current, after=head, insert=node)
        current = current.next

    return head


def nodes(head: Node, callback: Callable[[], Node] | None = None) -> Iterator[Node]:
    start_index: int = head.index

    current: Node = head
    steps: int = 0

    is_callback_callable: bool = isinstance(callback, Callable)

    while True:
        yield current
        current = current.next
        steps += 1

        is_node_in_list: bool = current.head is callback() if is_callback_callable else current.head is head.head

        if not is_node_in_list:
            if callback:
                head = callback()
                new_current: Node = head
            else:
                new_current: Node = head.head

            for _ in range(steps):
                new_current = new_current.next

            if new_current.index != steps:
                break

            current = new_current

        # TODO: index / step check could be more reliable than head in case callback or something else is broken
        if current is head or current.index == start_index:
            break


def is_consistent(node: Node) -> bool:
    is_correct: bool = True
    head: Node = node.head
    nodes__: list[Node] = []

    current: Node = head
    while True:
        nodes__.append(current)
        current = current.next
        if current is head:
            break

    for index, node in enumerate(nodes__):
        is_correct &= node.head is head
        is_correct &= node.index == index
        is_correct &= node.next.index == (node.index + 1) % len(nodes__)
        is_correct &= node.previous.index == (node.index - 1) % len(nodes__)
        is_correct &= node.next is nodes__[(index + 1) % len(nodes__)]
        is_correct &= node.previous is nodes__[(index - 1) % len(nodes__)]

        if not is_correct:
            break

    return is_correct


def length_step(node: Node):
    """
    Count nodes in circular doubly linked sequence by stepping through a full round.
    Necessary in contexts where Node indices are not guaranteed to be correct.
    """

    current: Node = node
    amount: int = 0

    while True:
        amount += 1
        current = current.next
        if current is node:
            break

    return amount


def length(node: Node) -> int:
    """
    Calculates Node amount in circular doubly linked list by index of last Node.
    Assumes Node indices are correct.
    """

    try:
        node.next.previous.value
    except AttributeError as exception:
        raise NotANodeError(f"'{node}' is not a Node, but a '{type(node)}'.") from exception

    return node.head.previous.index + 1


def remove(target: Node) -> Node:
    """
    Requires a minimum length of 2 nodes
    Assumes that target node is in a fully circular doubly linked list.
    """

    head: Node | None = None

    if target is target.head:
        head = target.next

    removed: Node = insert_between(before=target.previous, after=target.next, insert=None, head=head)

    return removed


def equal(first: Node, second: Node) -> bool:
    """Compare nodes from head in two circular doubly linked sequences."""
    equality: bool = True

    first_head: Node = first
    second_head: Node = second

    first_current: Node = first_head
    second_current: Node = second_head

    while True:
        equality &= first_current.value is second_current.value or first_current.value == second_current.value
        equality &= first_current.index == second_current.index
        equality &= first_current.head is first_head and second_current.head is second_head

        if not equality:
            break
        elif first_current.next is first_head and second_current.next is second_head:
            break
        elif first_current.next is first_head and second_current.next is not second_head:
            equality &= False
            break
        elif first_current.next is not first_head and second_current.next is second_head:
            equality &= False
            break

        first_current = first_current.next
        second_current = second_current.next

    return equality


def reverse_order(head: Node) -> Node:
    reversed_sequence: Node = head

    if head.next is not head:
        reversed_sequence: Node = head.previous
        previous_node: Node = head
        current_node: Node = head.previous

        while True:
            old_previous: Node = current_node.previous

            current_node.next = current_node.previous
            current_node.previous = previous_node

            previous_node = current_node
            current_node = old_previous

            if previous_node is head:
                break

    update_head(node=reversed_sequence)

    return reversed_sequence


def split_head_from_tail(node: Node) -> tuple[Node, Node | None]:
    first_head: Node = node

    if node.next is node:
        second_head: None = None
    else:
        second_head: Node = node.next
        split(first_head=first_head, second_head=second_head)

    return first_head, second_head


def middle_adjacent(head: Node) -> tuple[Node, Node]:
    """When node amount is uneven, preferentially adds one more node to before than after."""
    slow: Node = head
    fast: Node = head

    while fast.next is not head and fast.next.next is not head:
        slow = slow.next
        fast = fast.next.next

    before_last: Node = slow
    after_head: Node = slow.next

    return before_last, after_head


def stitch(head: Node, last: Node) -> None:
    """Stitch together head and last to make sequence circular."""
    head.previous, last.next = last, head


def split(first_head: Node, second_head: Node) -> None:
    """Split one circular doubly linked list into two."""

    if first_head.head is not second_head.head:
        raise NodesNotInSameListError(f"Nodes have different heads and are thus in different lists.")

    first_last: Node = second_head.previous
    second_last: Node = first_head.previous

    stitch(head=first_head, last=first_last), stitch(head=second_head, last=second_last)

    # TODO: refactor to extract repetitive code
    current: Node = first_head
    index: int = 0
    while True:
        current.head = first_head
        current.index = index

        current = current.next
        index += 1

        if current is first_head:
            break

    current: Node = second_head
    index: int = 0
    while True:
        current.head = second_head
        current.index = index

        current = current.next
        index += 1

        if current is second_head:
            break


def insert_between(before: Node, after: Node, insert: Node | None, head: Node | None = None) -> Node | None:
    """
    Requires a minimum length of 1 node
    When working with 1 node it is always going to assume that it should operate "between itself" and not on itself.
    Allows that insert can have multiple connected nodes.
    Assumes that each input node is in a fully circular doubly linked list.
    """

    if after.index - before.index == 0 and before.index != 0 and head is None:
        # When head is removed or overwritten and no new head has been defined, it is exception time!
        raise NoNewHeadError(f"Expected new head, but none was provided.")

    try:
        if insert is None:
            insert_head: Node = after
            insert_last: Node = before

            if length(node=before.head) == 1 or after.index - before.index == 1:
                processed_sequence_start: None = None
                processed_sequence_end: None = None
            else:
                processed_sequence_start: Node = before.next
                processed_sequence_end: Node = after.previous
        else:
            insert_head: Node = insert
            insert_last: Node = insert.previous

            if (after.index - before.index == 0 or after.index - before.index > 1) and length(node=before) > 1:
                processed_sequence_start: Node = before.next
                processed_sequence_end: Node = after.previous
            else:
                processed_sequence_start: None = None
                processed_sequence_end: None = None
    except AttributeError as exception:
        raise NotANodeError(f"'{insert}' is not a Node, but a '{type(insert)}'.") from exception

    if insert and processed_sequence_start and processed_sequence_end:
        stitch(head=insert_head, last=before), \
            stitch(head=after, last=insert_last), \
            stitch(head=processed_sequence_start, last=processed_sequence_end)
    elif not insert and processed_sequence_start and processed_sequence_end:
        stitch(head=insert_head, last=before), \
            stitch(head=processed_sequence_start, last=processed_sequence_end)
    elif insert:
        stitch(head=insert_head, last=before), \
            stitch(head=after, last=insert_last)
    elif not insert:
        pass
    else:
        raise ValueError(f"Unclear how Node sequences should be re-stitched...")

    # Update head and index in inserted nodes
    if head:
        update_head(node=head)
    elif insert or processed_sequence_start:
        # Update only after point of node insertion
        current: Node = insert_head
        if length_step(before) == 1:
            current_head: Node = current
        else:
            current_head: Node = current.previous.head

        if current is current_head:
            index: int = 0
        else:
            index: int = current.previous.index + 1

        while True:
            current.head = current_head
            current.index = index
            current = current.next
            if current is before.head:
                break
            index += 1

    if processed_sequence_start and processed_sequence_end:
        # Update removed nodes

        current_index = 0
        current_head = processed_sequence_start
        current: Node = processed_sequence_start
        while True:
            current.head = current_head
            current.index = current_index

            current = current.next

            if current is processed_sequence_start:
                break

            current_index += 1

    return processed_sequence_start


def update_head(node: Node) -> None:
    """Update all Nodes with new head."""
    if node is node.head and is_consistent(node=node):
        return

    current: Node = node
    index: int = 0

    while True:
        current.head = node
        current.index = index

        current = current.next
        index += 1

        if current is node:
            break


def before_target(current: Node, head: Node, target: Node) -> Node:
    """
    When last node value is still smaller than target value, then the last node is "before target".
    Assuming pre-sorted sub-lists.
    Assuming first head value is lower than or equal to insert head value.
    """

    while current.next.value <= target.value and current.next is not head:
        current = current.next

    return current


def split_in_middle(head: Node) -> tuple[Node, Node]:
    before_head: Node = head
    _, after_head = middle_adjacent(head=head)

    split(first_head=before_head, second_head=after_head)

    return before_head, after_head


def merge(first: Node, second: Node) -> Node:
    """
    Merge pre-sorted circular doubly linked nodes.
    Moving nodes from second into sorted positions in first.
    Assuming pre-sorted sub-lists.
    """

    if first.value > second.value:
        first, second = second, first

    before_insert: Node = first

    while second is not None:
        insert, second = split_head_from_tail(node=second)
        before_insert = before_target(current=before_insert, head=first, target=insert)
        insert_between(before=before_insert, after=before_insert.next, insert=insert)

    return first


def merge_sort(head: Node) -> Node:
    """Merge-sort implementation for circular doubly linked nodes."""

    if head.next is head:
        # When there is only one value, there is nothing to sort.
        merged_sorted: Node = head
    else:
        first, second = split_in_middle(head=head)
        first_sorted, second_sorted = merge_sort(head=first), merge_sort(head=second)
        merged_sorted: Node = merge(first=first_sorted, second=second_sorted)

    return merged_sorted



import pytest

from src.py_cdll import CDLL
from src.py_cdll.node import split_in_middle, Node, is_consistent


def test_split_in_middle_wrong_type_failure():
    # Setup
    node: str = "not a node"

    # Validation
    with pytest.raises(AttributeError):
        # noinspection PyTypeChecker
        split_in_middle(head=node)


def test_split_in_middle_cdll_one_node_success():
    # Setup
    datas0: list[int] = [0]
    cdll0: CDLL = CDLL(values=datas0)
    head0: Node = cdll0._head

    # Execution
    new_head, new_middle = split_in_middle(head=head0)

    # Validation
    assert new_head is head0
    assert new_head.next is head0
    assert new_head.previous is head0
    assert new_middle is head0
    assert new_middle.next is head0
    assert new_middle.previous is head0
    assert is_consistent(node=new_head) is True
    assert is_consistent(node=new_middle) is True


def test_split_in_middle_cdll_two_nodes_success():
    # Setup
    datas0: list[int] = [0, 1]
    cdll0: CDLL = CDLL(values=datas0)
    head0: Node = cdll0._head
    last0: Node = head0.next

    # Execution
    new_head, new_middle = split_in_middle(head=head0)

    # Validation
    assert new_head is head0
    assert new_head.next is head0
    assert new_head.previous is head0
    assert new_middle is last0
    assert new_middle.next is last0
    assert new_middle.previous is last0
    assert is_consistent(node=new_head) is True
    assert is_consistent(node=new_middle) is True


def test_split_in_middle_cdll_three_nodes_success():
    # Setup
    datas0: list[int] = [0, 1, 2]
    cdll0: CDLL = CDLL(values=datas0)
    head0: Node = cdll0._head
    middle0: Node = head0.next
    last0: Node = head0.previous

    # Execution
    new_head, new_middle = split_in_middle(head=head0)

    # Validation
    assert new_head is head0
    assert new_head.next is middle0
    assert new_head.previous is middle0
    assert new_middle is last0
    assert new_middle.next is last0
    assert new_middle.previous is last0
    assert is_consistent(node=new_head) is True
    assert is_consistent(node=new_middle) is True


def test_split_in_middle_cdll_four_nodes_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3]
    cdll0: CDLL = CDLL(values=datas0)
    head0: Node = cdll0._head
    second0: Node = head0.next
    third0: Node = second0.next
    last0: Node = third0.next

    # Execution
    new_head, new_middle = split_in_middle(head=head0)

    # Validation
    assert new_head is head0
    assert new_head.next is second0
    assert new_head.previous is second0
    assert new_middle is third0
    assert new_middle.next is last0
    assert new_middle.previous is last0
    assert is_consistent(node=new_head) is True
    assert is_consistent(node=new_middle) is True


def test_split_in_middle_cdll_seven_nodes_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6]
    cdll0: CDLL = CDLL(values=datas0)
    head0: Node = cdll0._head
    second0: Node = head0.next
    third0: Node = second0.next
    fourth0: Node = third0.next
    fifth0: Node = fourth0.next
    sixth0: Node = fifth0.next
    last0: Node = sixth0.next

    # Execution
    new_head, new_middle = split_in_middle(head=head0)

    # Validation
    assert new_head is head0
    assert new_head.next is second0
    assert new_head.previous is fourth0
    assert new_middle is fifth0
    assert new_middle.next is sixth0
    assert new_middle.previous is last0
    assert is_consistent(node=new_head) is True
    assert is_consistent(node=new_middle) is True



import pytest

from src.py_cdll import CDLL
from src.py_cdll.node import Node, merge, is_consistent


def test_merge_first_wrong_type_second_one_failure():
    # Setup
    first: str = "wrong type"
    datas1: list[int] = [0]
    cdll1: CDLL = CDLL(values=datas1)
    second: Node = cdll1._head

    # Validation
    with pytest.raises(AttributeError):
        # noinspection PyTypeChecker
        merge(first=first, second=second)


def test_merge_first_one_second_wrong_type_failure():
    # Setup
    datas0: list[int] = [0]
    cdll0: CDLL = CDLL(values=datas0)
    first: Node = cdll0._head
    second: str = "wrong type"

    # Validation
    with pytest.raises(AttributeError):
        # noinspection PyTypeChecker
        merge(first=first, second=second)


def test_merge_first_zero_second_zero_success():
    # Setup
    datas0: list[int] = [0]
    cdll0: CDLL = CDLL(values=datas0)
    first: Node = cdll0._head
    datas1: list[int] = [0]
    cdll1: CDLL = CDLL(values=datas1)
    second: Node = cdll1._head

    # Execution
    merged: Node = merge(first=first, second=second)

    # Validation
    assert merged is first
    assert merged.next is second
    assert merged.next.next is first
    assert merged.next.previous is first
    assert merged.previous is second
    assert merged.previous.next is first
    assert merged.previous.previous is first
    assert is_consistent(node=cdll0._head) is True
    assert is_consistent(node=cdll1._head) is True
    assert is_consistent(node=merged) is True


def test_merge_first_zero_second_one_success():
    # Setup
    datas0: list[int] = [0]
    cdll0: CDLL = CDLL(values=datas0)
    first: Node = cdll0._head
    datas1: list[int] = [1]
    cdll1: CDLL = CDLL(values=datas1)
    second: Node = cdll1._head

    # Execution
    merged: Node = merge(first=first, second=second)

    # Validation
    assert merged is first
    assert merged.next is second
    assert merged.next.next is first
    assert merged.next.previous is first
    assert merged.previous is second
    assert merged.previous.next is first
    assert merged.previous.previous is first
    assert is_consistent(node=cdll0._head) is True
    assert is_consistent(node=cdll1._head) is True
    assert is_consistent(node=merged) is True


def test_merge_first_one_second_zero_success():
    # Setup
    datas0: list[int] = [1]
    cdll0: CDLL = CDLL(values=datas0)
    first: Node = cdll0._head
    datas1: list[int] = [0]
    cdll1: CDLL = CDLL(values=datas1)
    second: Node = cdll1._head

    # Execution
    merged: Node = merge(first=first, second=second)

    # Validation
    assert merged is second
    assert merged.next is first
    assert merged.next.next is second
    assert merged.next.previous is second
    assert merged.previous is first
    assert merged.previous.next is second
    assert merged.previous.previous is second
    assert is_consistent(node=cdll0._head) is True
    assert is_consistent(node=cdll1._head) is True
    assert is_consistent(node=merged) is True


def test_merge_first_zero_and_two_second_one_success():
    # Setup
    datas0: list[int] = [0, 2]
    cdll0: CDLL = CDLL(values=datas0)
    first0: Node = cdll0._head
    first1: Node = first0.next
    datas1: list[int] = [1]
    cdll1: CDLL = CDLL(values=datas1)
    second0: Node = cdll1._head

    # Execution
    merged: Node = merge(first=first0, second=second0)

    # Validation
    assert merged is first0
    assert merged.next is second0
    assert merged.previous is first1

    assert merged.next.next is first1
    assert merged.next.previous is first0

    assert merged.next.next.next is first0
    assert merged.next.next.previous is second0

    assert merged.previous.next is first0
    assert merged.previous.previous is second0

    assert is_consistent(node=cdll0._head) is True
    assert is_consistent(node=cdll1._head) is True
    assert is_consistent(node=merged) is True


def test_merge_first_five_second_two_and_seven_success():
    # Setup
    datas0: list[int] = [5]
    cdll0: CDLL = CDLL(values=datas0)
    first0: Node = cdll0._head
    datas1: list[int] = [2, 7]
    cdll1: CDLL = CDLL(values=datas1)
    second0: Node = cdll1._head
    second1: Node = second0.next

    # Execution
    merged: Node = merge(first=first0, second=second0)

    # Validation
    assert merged is second0
    assert merged.next is first0
    assert merged.previous is second1

    assert merged.next.next is second1
    assert merged.next.previous is second0

    assert merged.next.next.next is second0
    assert merged.next.next.previous is first0

    assert merged.previous.next is second0
    assert merged.previous.previous is first0

    assert is_consistent(node=cdll0._head) is True
    assert is_consistent(node=cdll1._head) is True
    assert is_consistent(node=merged) is True


def test_merge_first_seven_and_thirteen_second_three_and_sixteen_success():
    # Setup
    datas0: list[int] = [7, 13]
    cdll0: CDLL = CDLL(values=datas0)
    first0: Node = cdll0._head
    first1: Node = first0.next
    datas1: list[int] = [3, 16]
    cdll1: CDLL = CDLL(values=datas1)
    second0: Node = cdll1._head
    second1: Node = second0.next

    # Execution
    merged: Node = merge(first=first0, second=second0)

    # Validation
    assert merged is second0
    assert merged.next is first0
    assert merged.previous is second1

    assert merged.next.next is first1
    assert merged.next.previous is second0

    assert merged.next.next.next is second1
    assert merged.next.next.previous is first0

    assert merged.next.next.next.next is second0
    assert merged.next.next.next.previous is first1

    assert merged.previous.next is second0
    assert merged.previous.previous is first1

    assert is_consistent(node=cdll0._head) is True
    assert is_consistent(node=cdll1._head) is True
    assert is_consistent(node=merged) is True



import pytest

from src.py_cdll import CDLL, EmptyCDLLError, UnevenListLengthError, ValueNotFoundError, MultipleValuesFoundError


def test_opposite_two_items_from_zero_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"

    list0: CDLL = CDLL()
    list0.append(value=data0)
    list0.append(value=data1)

    # Execution
    data2: str = list0.opposite_unique(value=data0)

    # Validation
    assert data2 == data1


def test_opposite_four_items_from_one_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"

    list0: CDLL = CDLL()
    list0.append(value=data0)
    list0.append(value=data1)
    list0.append(value=data2)
    list0.append(value=data3)

    # Execution
    data4: str = list0.opposite_unique(value=data1)

    # Validation
    assert data4 == data3


def test_opposite_sixteen_items_from_thirteen_across_zero_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    data7: str = "data7"
    data8: str = "data8"
    data9: str = "data9"
    data10: str = "data10"
    data11: str = "data11"
    data12: str = "data12"
    data13: str = "data13"
    data14: str = "data14"
    data15: str = "data15"

    list0: CDLL = CDLL()
    list0.append(value=data0)
    list0.append(value=data1)
    list0.append(value=data2)
    list0.append(value=data3)
    list0.append(value=data4)
    list0.append(value=data5)
    list0.append(value=data6)
    list0.append(value=data7)
    list0.append(value=data8)
    list0.append(value=data9)
    list0.append(value=data10)
    list0.append(value=data11)
    list0.append(value=data12)
    list0.append(value=data13)
    list0.append(value=data14)
    list0.append(value=data15)

    # Execution
    data16: str = list0.opposite_unique(value=data13)

    # Validation
    assert data16 == data5


def test_opposite_empty_list_data_not_found_error_failure():
    # Setup
    data0: str = "data0"

    list0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        list0.opposite_unique(value=data0)


def test_opposite_one_item_failure():
    # Setup
    data0: str = "data0"

    list0: CDLL = CDLL()
    list0.append(value=data0)

    # Validation
    with pytest.raises(UnevenListLengthError):
        list0.opposite_unique(value=data0)


def test_opposite_five_items_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data0"
    data2: str = "data0"
    data3: str = "data0"
    data4: str = "data0"

    list0: CDLL = CDLL()
    list0.append(value=data0)
    list0.append(value=data1)
    list0.append(value=data2)
    list0.append(value=data3)
    list0.append(value=data4)

    # Validation
    with pytest.raises(UnevenListLengthError):
        list0.opposite_unique(value=data1)


def test_opposite_thirteen_items_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data0"
    data2: str = "data0"
    data3: str = "data0"
    data4: str = "data0"
    data5: str = "data0"
    data6: str = "data0"
    data7: str = "data0"
    data8: str = "data0"
    data9: str = "data0"
    data10: str = "data0"
    data11: str = "data0"
    data12: str = "data0"

    list0: CDLL = CDLL()
    list0.append(value=data0)
    list0.append(value=data1)
    list0.append(value=data2)
    list0.append(value=data3)
    list0.append(value=data4)
    list0.append(value=data5)
    list0.append(value=data6)
    list0.append(value=data7)
    list0.append(value=data8)
    list0.append(value=data9)
    list0.append(value=data10)
    list0.append(value=data11)
    list0.append(value=data12)

    # Validation
    with pytest.raises(UnevenListLengthError):
        list0.opposite_unique(value=data9)


def test_opposite_six_items_data_not_found_failure():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"

    list0: CDLL = CDLL()
    list0.append(value=data0)
    list0.append(value=data1)
    list0.append(value=data2)
    list0.append(value=data3)
    list0.append(value=data4)
    list0.append(value=data5)

    # Validation
    with pytest.raises(ValueNotFoundError):
        list0.opposite_unique(value=data)


def test_opposite_six_items_multiple_data_instances_found_failure():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data"
    data5: str = "data5"

    list0: CDLL = CDLL()
    list0.append(value=data0)
    list0.append(value=data1)
    list0.append(value=data2)
    list0.append(value=data3)
    list0.append(value=data4)
    list0.append(value=data5)

    # Validation
    with pytest.raises(MultipleValuesFoundError):
        list0.opposite_unique(value=data)



import pytest

from src.py_cdll import CDLL
from src.py_cdll.node import merge_sort, nodes_from_values, Node, equal, is_consistent


def test_merge_sort_integers_success():
    # Setup
    values0 = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    values1 = [2, 0, 9, 7, 4, 5, 6, 3, 8, 1]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = nodes_from_values(values=values1)

    # Execution
    node2: Node = merge_sort(head=node1)

    # Validation
    assert equal(first=node0, second=node2)
    assert is_consistent(node=node2) is True


def test_merge_sort_strings_success():
    # Setup
    values0 = ["a", "d", "f", "j", "m", "p", "q", "t", "v", "y"]
    values1 = ['t', 'v', 'y', 'j', 'm', 'a', 'p', 'q', 'd', 'f']
    node0: Node = nodes_from_values(values=values0)
    node1: Node = nodes_from_values(values=values1)

    # Execution
    node2: Node = merge_sort(head=node1)

    # Validation
    assert equal(first=node0, second=node2)
    assert is_consistent(node=node2) is True


def test_merge_sort_functions_failure():
    # Setup
    l0 = [min, max, all]
    c0 = CDLL(l0)

    # Validation
    with pytest.raises(TypeError):
        merge_sort(head=c0._head)



from src.py_cdll import CDLL


def test___eq___other_type_success():
    # Setup
    cdll0: CDLL = CDLL()
    other0: float = 3.14

    # Execution
    # noinspection PyTypeChecker
    equal: bool = cdll0 == other0

    # Validation
    assert equal is False


def test___eq___empty_and_empty_success():
    # Setup
    cdll0: CDLL = CDLL()
    cdll1: CDLL = CDLL()

    # Execution
    equal: bool = cdll0 == cdll1

    # Validation
    assert equal is True


def test___eq___one_and_one_success():
    # Setup
    data0: int = 0
    data1: int = 0
    datas0: list[int] = [data0]
    datas1: list[int] = [data1]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)

    # Execution
    equal: bool = cdll0 == cdll1

    # Validation
    assert equal is True


def test___eq___six_and_six_success():
    # Setup
    data00: int = 0
    data01: int = 1
    data02: int = 2
    data03: int = 3
    data04: int = 4
    data05: int = 5
    data10: int = 0
    data11: int = 1
    data12: int = 2
    data13: int = 3
    data14: int = 4
    data15: int = 5
    datas0: list[int] = [data00, data01, data02, data03, data04, data05]
    datas1: list[int] = [data10, data11, data12, data13, data14, data15]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)

    # Execution
    equal: bool = cdll0 == cdll1

    # Validation
    assert equal is True


def test___eq___one_and_zero_success():
    # Setup
    data0: int = 0
    datas0: list[int] = [data0]
    datas1: list[int] = []
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)

    # Execution
    equal: bool = cdll0 == cdll1

    # Validation
    assert equal is False


def test___eq___zero_and_one_success():
    # Setup
    data0: int = 0
    datas0: list[int] = []
    datas1: list[int] = [data0]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)

    # Execution
    equal: bool = cdll0 == cdll1

    # Validation
    assert equal is False


def test___eq___four_and_four_last_different_success():
    # Setup
    data00: int = 0
    data01: int = 1
    data02: int = 2
    data03: int = 3
    data10: int = 0
    data11: int = 1
    data12: int = 2
    data13: int = 33
    datas0: list[int] = [data00, data01, data02, data03]
    datas1: list[int] = [data10, data11, data12, data13]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)

    # Execution
    equal: bool = cdll0 == cdll1

    # Validation
    assert equal is False



from src.py_cdll import CDLL


def test_copy_empty_success():
    # Setup
    cdll0: CDLL = CDLL()

    # Execution
    cdll1: CDLL = cdll0.copy()

    # Validation
    assert cdll0 is not cdll1
    assert cdll0 == cdll1


def test_copy_single_value_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll1: CDLL = cdll0.copy()

    # Validation
    assert cdll0 is not cdll1
    assert cdll0 == cdll1


def test_copy_multiple_values_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll1: CDLL = cdll0.copy()

    # Validation
    assert cdll0 is not cdll1
    assert cdll0 == cdll1



import pytest

from src.py_cdll import CDLL


def test_index_with_empty_list_failure():
    # Setup
    data0: str = "data0"
    list0: CDLL = CDLL()

    # Validation
    with pytest.raises(ValueError):
        list0.index(value=data0)


def test_index_single_element_single_hit_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    list0: CDLL = CDLL(values=datas0)

    # Execution
    index0: int = list0.index(value=data0)

    # Validation
    assert index0 == 0


def test_index_multiple_elements_single_hit_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    list0: CDLL = CDLL(values=datas0)

    # Execution
    index0: int = list0.index(value=data0)

    # Validation
    assert index0 == 0


def test_index_multiple_elements_no_hits_failure():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    list0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        list0.index(value=data)


def test_index_multiple_elements_multiple_hits_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data, data1, data, data2]
    list0: CDLL = CDLL(values=datas0)

    # Execution
    index0: int = list0.index(value=data)

    # Validation
    assert index0 == 1


def test_index_found_value_stop_out_of_range_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index0: int = cdll0.index(value=data1, stop=len(cdll0) + 1)

    # Validation
    assert index0 == 1


def test_index_not_found_value_stop_out_of_range_failure():
    # Setup
    value0: str = "data5"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0.index(value=value0, stop=len(cdll0) + 5)


def test_index_found_value_in_range_stop_mid_list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index0: int = cdll0.index(value=data2, stop=len(cdll0) // 2)

    # Validation
    assert index0 == 2


def test_index_not_found_value_out_of_range_stop_mid_list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0.index(value=data4, stop=len(cdll0) // 2)


def test_index_not_found_value_unavailable_stop_mid_list_success():
    # Setup
    value0: str = "data9"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0.index(value=value0, stop=len(cdll0) // 2)


def test_index_found_value_in_range_start_mid_list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index0: int = cdll0.index(value=data5, start=len(cdll0) // 2)

    # Validation
    assert index0 == 5


def test_index_not_found_value_out_of_range_start_mid_list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0.index(value=data2, start=len(cdll0) // 2)


def test_index_not_found_value_unavailable_start_mid_list_success():
    # Setup
    value0: str = "data9"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0.index(value=value0, start=len(cdll0) // 2)


def test_index_not_found_value_out_of_range_slice_zero_length_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0.index(value=data1, start=1, stop=1)


def test_index_not_found_value_unavailable_slice_zero_length_success():
    # Setup
    value0: str = "data6"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    with pytest.raises(ValueError):
        cdll0.index(value=value0, start=2, stop=2)


def test_index_value_found_slice_start_negative_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index0: int = cdll0.index(value=data2, start=-3, stop=3)

    # Validation
    assert index0 == 2


def test_index_value_found_slice_stop_negative_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index0: int = cdll0.index(value=data2, start=2, stop=-2)

    # Validation
    assert index0 == 2


def test_index_value_found_slice_both_negative_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    index0: int = cdll0.index(value=data2, start=-3, stop=-2)

    # Validation
    assert index0 == 2



import pytest

from src.py_cdll import CDLL


def test__wraparound_index_length_zero_index_one_failure():
    # Setup
    datas0: list = []
    cdll0: CDLL = CDLL(values=datas0)
    index0: int = 1

    # Validation
    with pytest.raises(ZeroDivisionError):
        cdll0._wraparound_index(index=index0)


def test__wraparound_index_length_one_index_six_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL[str] = CDLL(values=datas0)
    index0: int = 6
    index1: int = 0

    # Execution
    index2: int = cdll0._wraparound_index(index=index0)

    # Validation
    assert index2 == index1


def test__wraparound_index_length_two_index_thirty_seven_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    datas0: list[str] = [data0, data1]
    cdll0: CDLL[str] = CDLL(values=datas0)
    index0: int = 37
    index1: int = 1

    # Execution
    index2: int = cdll0._wraparound_index(index=index0)

    # Validation
    assert index2 == index1


def test__wraparound_index_length_three_index_minus_five_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL[str] = CDLL(values=datas0)
    index0: int = -5
    index1: int = 1

    # Execution
    index2: int = cdll0._wraparound_index(index=index0)

    # Validation
    assert index2 == index1


def test__wraparound_index_length_four_index_minus_one_hundred_twenty_seven_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data0, data1, data2, data3]
    cdll0: CDLL[str] = CDLL(values=datas0)
    index0: int = -127
    index1: int = 1

    # Execution
    index2: int = cdll0._wraparound_index(index=index0)

    # Validation
    assert index2 == index1


def test__wraparound_index_length_five_index_ten_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL[str] = CDLL(values=datas0)
    index0: int = 10
    index1: int = 0

    # Execution
    index2: int = cdll0._wraparound_index(index=index0)

    # Validation
    assert index2 == index1



import pytest

from src.py_cdll import CDLL


def test__normalize_index_value_zero_success():
    # Setup
    values0: range = range(5)
    cdll0: CDLL = CDLL()
    [cdll0.append(value) for value in values0]
    index0: int = 0

    # Execution
    index1: int = cdll0._normalize_index(index=index0)

    # Validation
    assert index0 == index1


def test__normalize_index_value_positive_success():
    # Setup
    values0: range = range(8)
    cdll0: CDLL = CDLL()
    [cdll0.append(value) for value in values0]
    index0: int = 4

    # Execution
    index1: int = cdll0._normalize_index(index=index0)

    # Validation
    assert index0 == index1


def test__normalize_index_value_negative_success():
    # Setup
    values0: range = range(11)
    cdll0: CDLL = CDLL()
    [cdll0.append(value) for value in values0]
    index0: int = -6
    index1: int = 5

    # Execution
    index2: int = cdll0._normalize_index(index=index0)

    # Validation
    assert index1 == index2


def test__normalize_index_value_positive_out_of_range_failure():
    # Setup
    values0: range = range(20)
    cdll0: CDLL = CDLL()
    [cdll0.append(value) for value in values0]
    index0: int = 25

    # Validation
    with pytest.raises(IndexError):
        cdll0._normalize_index(index=index0)


def test__normalize_index_value_negative_out_of_range_failure():
    # Setup
    values0: range = range(37)
    cdll0: CDLL = CDLL()
    [cdll0.append(value) for value in values0]
    index0: int = -45

    # Validation
    with pytest.raises(IndexError):
        cdll0._normalize_index(index=index0)



import pytest

from src.py_cdll import CDLL


def test___getitem___slice_single_item_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [3]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(2, 3, 1)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_multiple_items_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [2, 3, 4]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(1, 4, None)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_single_item_last_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(4, 5, None)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_multiple_from_head_through_init_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [4, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(3, 5, None)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_multiple_items_spread_out_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas1: list[int] = [2, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(2, 7, 3)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_multiple_items_spread_out_less_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas1: list[int] = [0, 2, 4, 6, 8]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(0, None, 2)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_all_items_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 4, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(0, None, None)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_all_items_away_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = []
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(5, None, None)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_no_items_with_index_out_of_range_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = []
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(37, None, None)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_zero_step_failure():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    slice0: slice = slice(None, None, 0)

    # Validation
    with pytest.raises(ValueError):
        _ = cdll0[slice0]


def test___getitem___slice_negative_two_step_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [4, 2]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(3, None, -2)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_out_of_range_step_from_zero_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(0, None, 11)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1


def test___getitem___slice_out_of_range_step_from_out_of_range_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = []
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(37, None, 11)

    # Execution
    cdll2: CDLL[int] = cdll0[slice0]

    # Validation
    assert cdll2 == cdll1



import pytest

from src.py_cdll import CDLL


def test___add___wrong_type_failure():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)
    adder0: int = 22

    # Validation
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        cdll0 + adder0


def test___add___list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=[])
    cdll2: CDLL = CDLL(values=datas0 + datas0)

    # Execution
    cdll3: CDLL = cdll0 + datas0

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___add___empty_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=[])
    cdll2: CDLL = CDLL(values=datas0)

    # Execution
    cdll3: CDLL = cdll0 + cdll1

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___add___one_value_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)

    # Execution
    cdll3: CDLL = cdll0 + cdll1

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___add___two_values_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)

    # Execution
    cdll3: CDLL = cdll0 + cdll1

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___add___five_values_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    data7: str = "data7"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3, data4, data5, data6, data7]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)

    # Execution
    cdll3: CDLL = cdll0 + cdll1

    # Validation
    assert cdll3 == cdll2
    assert cdll3 is not cdll0 and cdll3 is not cdll1 and cdll3 is not cdll2


def test___add___item_and_empty_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    datas1: list[str] = []
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0)

    # Execution
    cdll3: CDLL = cdll0 + cdll1

    # Validation
    assert cdll3 == cdll2


def test___add___empty_and_item_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = []
    datas1: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas1)

    # Execution
    cdll3: CDLL = cdll0 + cdll1

    # Validation
    assert cdll3 == cdll2


def test___add___item_and_item_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    datas0: list[str] = [data0]
    datas1: list[str] = [data1]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)

    # Execution
    cdll3: CDLL = cdll0 + cdll1

    # Validation
    assert cdll3 == cdll2


def test___add___two_items_and_five_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1]
    datas1: list[str] = [data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)

    # Execution
    cdll3: CDLL = cdll0 + cdll1

    # Validation
    assert cdll3 == cdll2


def test___add___three_items_and_four_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)

    # Execution
    cdll3: CDLL = cdll0 + cdll1

    # Validation
    assert cdll3 == cdll2


def test___add___cdll_tuple_success():
    # Setup
    cdll0: CDLL = CDLL()
    tuple1: tuple = tuple()

    # Execution
    cdll1: CDLL = cdll0 + tuple1

    # Validation
    assert cdll1 == cdll0


def test___add___cdll_list_success():
    # Setup
    cdll0: CDLL = CDLL()
    list1: list = list()

    # Execution
    cdll1: CDLL = cdll0 + list1

    # Validation
    assert cdll1 == cdll0


def test___add___cdll_set_success():
    # Setup
    cdll0: CDLL = CDLL()
    set1: set = set()

    # Execution
    cdll1: CDLL = cdll0 + set1

    # Validation
    assert cdll1 == cdll0


def test___add___cdll_range_success():
    # Setup
    cdll0: CDLL = CDLL()
    cdll1: CDLL = CDLL(values=[0, 1, 2, 3, 4])
    range1: range = range(5)

    # Execution
    cdll2: CDLL = cdll0 + range1

    # Validation
    assert cdll2 == cdll1



from src.py_cdll import CDLL


def test__indices_from_slice_first_last_single_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    cdll0: CDLL = CDLL(values=datas0)
    slice0: slice = slice(0, 10, 1)

    # Execution
    indices0: list[int] = list(cdll0._range_from_slice(segment=slice0))

    # Validation
    assert indices0 == datas0


def test__indices_from_slice_first_none_double_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas1: list[int] = [0, 2, 4, 6, 8]
    cdll0: CDLL = CDLL(values=datas0)
    slice0: slice = slice(0, None, 2)

    # Execution
    indices0: list[int] = list(cdll0._range_from_slice(segment=slice0))

    # Validation
    assert indices0 == datas1


def test__indices_from_slice_first_none_triple_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas1: list[int] = [0, 3, 6, 9]
    cdll0: CDLL = CDLL(values=datas0)
    slice0: slice = slice(0, None, 3)

    # Execution
    indices0: list[int] = list(cdll0._range_from_slice(segment=slice0))

    # Validation
    assert indices0 == datas1


def test__indices_from_slice_second_penultimate_quad_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas1: list[int] = [1, 5]
    cdll0: CDLL = CDLL(values=datas0)
    slice0: slice = slice(1, 9, 4)

    # Execution
    indices0: list[int] = list(cdll0._range_from_slice(segment=slice0))

    # Validation
    assert indices0 == datas1


def test__indices_from_slice_last_none_none_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas1: list[int] = []
    cdll0: CDLL = CDLL(values=datas0)
    slice0: slice = slice(10, None, None)

    # Execution
    indices0: list[int] = list(cdll0._range_from_slice(segment=slice0))

    # Validation
    assert indices0 == datas1



from collections.abc import Iterator

import pytest

from src.py_cdll.node import Node, nodes_from_values, nodes, remove, insert_between, is_consistent


def test_nodes_five_items_success():
    # Setup
    head0: Node = nodes_from_values(values=[0, 1, 2, 3, 4])
    iterator0: Iterator[Node] = nodes(head=head0)

    # Execution
    node0: Node = next(iterator0)
    node1: Node = next(iterator0)
    node2: Node = next(iterator0)
    node3: Node = next(iterator0)
    node4: Node = next(iterator0)

    # Validation
    assert head0 is node0
    assert head0.next is node1
    assert head0.next.next is node2
    assert head0.next.next.next is node3
    assert head0.next.next.next.next is node4
    assert head0.previous is node4
    assert head0.previous.previous is node3
    assert head0.previous.previous.previous is node2
    assert head0.previous.previous.previous.previous is node1
    assert head0.previous.previous.previous.previous.previous is node0
    assert is_consistent(node=head0)


def test_nodes_five_items_middle_removed_success():
    # Setup
    head0: Node = nodes_from_values(values=[0, 1, 2, 3, 4])
    node2: Node = head0.next.next
    iterator0: Iterator[Node] = nodes(head=head0)

    # Execution
    remove(target=node2)
    node0: Node = next(iterator0)
    node1: Node = next(iterator0)
    node3: Node = next(iterator0)
    node4: Node = next(iterator0)

    # Validation
    with pytest.raises(StopIteration):
        _ = next(iterator0)

    assert head0 is node0
    assert head0.next is node1
    assert head0.next.next is node3
    assert head0.next.next.next is node4
    assert head0.previous is node4
    assert head0.previous.previous is node3
    assert head0.previous.previous.previous is node1
    assert head0.previous.previous.previous.previous is node0
    assert is_consistent(node=head0)


def test_nodes_five_items_three_steps_remove_three_failure():
    # Setup
    head0: Node = nodes_from_values(values=[0, 1, 2, 3, 4])
    node2: Node = head0.next.next
    node3: Node = head0.next.next.next
    node4: Node = head0.next.next.next.next
    iterator0: Iterator[Node] = nodes(head=head0)

    # Execution
    node0: Node = next(iterator0)
    node1: Node = next(iterator0)
    _: Node = next(iterator0)
    _: Node = remove(target=node2)
    _: Node = remove(target=node3)
    _: Node = remove(target=node4)

    # Validation
    with pytest.raises(StopIteration):
        _ = next(iterator0)

    assert head0 is node0
    assert head0.next is node1
    assert head0.previous is node1
    assert head0.previous.previous is node0
    assert is_consistent(node=head0)


def test_nodes_five_items_three_steps_pop_four_append_five_success():
    # Setup
    head0: Node = nodes_from_values(values=[0, 1, 2, 3, 4])
    head1: Node = nodes_from_values(values=[5, 6, 7, 8, 9])
    node3: Node = head0.next.next.next
    node4: Node = head0.next.next.next.next
    iterator0: Iterator[Node] = nodes(head=head0)

    # Execution
    node0: Node = next(iterator0)
    node1: Node = next(iterator0)
    node2: Node = next(iterator0)
    _: Node = remove(target=node1)
    _: Node = remove(target=node2)
    _: Node = remove(target=node3)
    _: Node = remove(target=node4)
    insert_between(before=node0, after=node0, insert=head1)
    node3: Node = next(iterator0)
    node4: Node = next(iterator0)
    node5: Node = next(iterator0)

    # Validation
    with pytest.raises(StopIteration):
        _ = next(iterator0)

    assert head0 is node0
    assert head0.next is head1
    assert head0.next.next is head1.next
    assert head0.next.next.next is node3
    assert head0.next.next.next.next is node4
    assert head0.next.next.next.next.next is node5
    assert head0.previous is node5
    assert head0.previous.previous is node4
    assert head0.previous.previous.previous is node3
    assert head0.previous.previous.previous.previous is head1.next
    assert head0.previous.previous.previous.previous.previous is head1
    assert is_consistent(node=head0)
    assert is_consistent(node=head1)


def test_nodes_with_callback_five_items_three_steps_pop_four_append_five_success():
    # Setup
    head0: Node = nodes_from_values(values=[0, 1, 2, 3, 4])
    head1: Node = nodes_from_values(values=[5, 6, 7, 8, 9])
    node3: Node = head0.next.next.next
    node4: Node = head0.next.next.next.next
    callback0: Node = head0
    iterator0: Iterator[Node] = nodes(head=head0, callback=lambda: callback0)

    # Execution
    node0: Node = next(iterator0)
    node1: Node = next(iterator0)
    node2: Node = next(iterator0)
    _: Node = remove(target=node1)
    _: Node = remove(target=node2)
    _: Node = remove(target=node3)
    _: Node = remove(target=node4)
    insert_between(before=node0, after=node0, insert=head1, head=head1)
    callback0: Node = head1
    node3: Node = next(iterator0)
    node4: Node = next(iterator0)
    node5: Node = next(iterator0)

    # Validation
    with pytest.raises(StopIteration):
        _ = next(iterator0)

    assert head0 is node0
    assert head0.next is head1
    assert head0.next.next is head1.next
    assert head0.next.next.next is head1.next.next
    assert head0.next.next.next.next is node3
    assert head0.next.next.next.next.next is node4
    assert head0.next.next.next.next.next.next is node5
    assert head0.previous is node4
    assert head0.previous.previous is node3
    assert head0.previous.previous.previous is head1.next.next
    assert head0.previous.previous.previous.previous is head1.next
    assert head0.previous.previous.previous.previous.previous is head1
    assert is_consistent(node=head0)
    assert is_consistent(node=head1)



import pytest

from src.py_cdll import CDLL, EmptyCDLLError, ValueNotFoundError, MultipleValuesFoundError


def test_index_unique_with_empty_list_failure():
    # Setup
    data0 = "data0"
    list0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        list0.index_unique(data0)


def test_index_unique_with_single_option_single_hit_success():
    # Setup
    data0 = "data0"
    list0: CDLL = CDLL()
    list0.append(data0)

    # Validation
    assert list0.index_unique(data0) == 0


def test_index_unique_with_multiple_options_single_hit_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)

    # Validation
    assert list0.index_unique(data1) == 1


def test_index_unique_with_zero_hit_failure():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)

    # Validation
    with pytest.raises(ValueNotFoundError):
        list0.index_unique(data)


def test_index_unique_with_multiple_hits_failure():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    list0: CDLL = CDLL()
    list0.append(data)
    list0.append(data0)
    list0.append(data)
    list0.append(data1)
    list0.append(data)
    list0.append(data2)
    list0.append(data)

    # Validation
    with pytest.raises(MultipleValuesFoundError):
        list0.index_unique(data)



from src.py_cdll import CDLL


def test_clear_empty_list_success():
    # Setup
    cdll0: CDLL = CDLL()

    # Execution
    cdll0.clear()

    # Validation
    assert len(cdll0) == 0
    assert cdll0._head is None


def test_clear_list_with_content_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0.clear()

    # Validation
    assert len(cdll0) == 0
    assert cdll0._head is None



from src.py_cdll.node import Node, insert_between, is_consistent


def test_head_one_node_success():
    # Setup
    node0: Node = Node(value="zero")

    # Validation
    assert node0.head is node0
    assert is_consistent(node=node0) is True


def test_head_two_nodes_insert_sequentially_success():
    # Setup
    node0: Node = Node(value="zero")
    node1: Node = Node(value="one")
    insert_between(before=node0, after=node0, insert=node1)

    # Validation
    assert node0.head is node0
    assert node1.head is node0
    assert is_consistent(node=node0) is True


def test_head_three_nodes_insert_sequentially_success():
    # Setup
    node0: Node = Node(value="zero")
    node1: Node = Node(value="one")
    node2: Node = Node(value="two")
    insert_between(before=node1, after=node1, insert=node2)
    insert_between(before=node0, after=node0, insert=node1)

    # Validation
    assert node0.head is node0
    assert node1.head is node0
    assert node2.head is node0
    assert is_consistent(node=node0) is True


def test_head_four_nodes_insert_sequentially_and_then_overwrite_head_success():
    # Setup
    node0: Node = Node(value="zero")
    node1: Node = Node(value="one")
    node2: Node = Node(value="two")
    node3: Node = Node(value="three")
    insert_between(before=node1, after=node1, insert=node2)
    insert_between(before=node0, after=node0, insert=node1)
    insert_between(before=node2, after=node0, insert=node3, head=node3)

    # Validation
    assert node0.head is node3
    assert node1.head is node3
    assert node2.head is node3
    assert node3.head is node3
    assert is_consistent(node=node0) is True



from src.py_cdll.node import Node, reverse_order, is_consistent, nodes_from_values


def test_reverse_order_one_node_success():
    # Setup
    node0: Node = nodes_from_values(values=["zero"])

    # Execution
    reversed0: Node = reverse_order(head=node0)

    # Validation
    assert reversed0 is node0
    assert reversed0.head is node0
    assert reversed0.index == 0
    assert reversed0.next is node0
    assert reversed0.previous is node0
    assert is_consistent(node=reversed0) is True


def test_reverse_order_two_node_success():
    # Setup
    values0: list[str] = ["zero", "one"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = node0.next

    # Execution
    reversed0: Node = reverse_order(head=node0)

    # Validation
    assert reversed0 is node1
    assert reversed0.head is node1
    assert reversed0.index == 0
    assert reversed0.next is node0
    assert reversed0.next.head is node1
    assert reversed0.next.index == 1
    assert reversed0.previous is node0
    assert is_consistent(node=reversed0) is True


def test_reverse_order_five_nodes_success():
    # Setup
    node0: Node = nodes_from_values(values=["zero", "one", "two", "three", "four"])
    node1: Node = node0.next
    node2: Node = node0.next.next
    node3: Node = node0.next.next.next
    node4: Node = node0.next.next.next.next

    # Execution
    reversed0: Node = reverse_order(head=node0)

    # Validation
    assert reversed0 is node4
    assert reversed0.next is node3
    assert reversed0.next.next is node2
    assert reversed0.next.next.next is node1
    assert reversed0.next.next.next.next is node0
    assert reversed0.next.next.next.next.next is node4
    assert reversed0.previous is node0
    assert reversed0.previous.previous is node1
    assert reversed0.previous.previous.previous is node2
    assert reversed0.previous.previous.previous.previous is node3
    assert reversed0.previous.previous.previous.previous.previous is node4
    assert reversed0.index == 0
    assert reversed0.next.index == 1
    assert reversed0.next.next.index == 2
    assert reversed0.next.next.next.index == 3
    assert reversed0.next.next.next.next.index == 4
    assert reversed0.next.next.next.next.next.index == 0
    assert reversed0.head is node4
    assert reversed0.next.head is node4
    assert reversed0.next.next.head is node4
    assert reversed0.next.next.next.head is node4
    assert reversed0.next.next.next.next.head is node4
    assert reversed0.next.next.next.next.next.head is node4
    assert is_consistent(node=reversed0) is True



from src.py_cdll.node import Node, remove, nodes_from_values, is_consistent


def test_remove_first_in_two_node_sequence_success():
    # Setup
    head0: Node = nodes_from_values(values=["zero", "one"])
    node0: Node = head0
    node1: Node = head0.next

    # Execution
    removed0: Node = remove(target=node0)

    # Validation
    assert removed0 is node0
    assert node0.next is node0
    assert node0.previous is node0
    assert node1.next is node1
    assert node1.previous is node1
    assert is_consistent(node=node0) is True
    assert is_consistent(node=node1) is True
    assert is_consistent(node=removed0) is True


def test_remove_last_in_three_node_sequence_success():
    # Setup
    head0: Node = nodes_from_values(values=["zero", "one", "two"])
    node0: Node = head0
    node1: Node = head0.next
    node2: Node = head0.next.next

    # Execution
    removed0: Node = remove(target=node2)

    # Validation
    assert removed0 is node2
    assert node0.next is node1
    assert node0.previous is node1
    assert node1.next is node0
    assert node1.previous is node0
    assert node2.next is node2
    assert node2.previous is node2
    assert is_consistent(node=node0) is True
    assert is_consistent(node=node1) is True
    assert is_consistent(node=node2) is True
    assert is_consistent(node=removed0) is True


def test_remove_middle_in_five_node_sequence_success():
    # Setup
    head0: Node = nodes_from_values(values=["zero", "one", "two", "three", "four"])
    node0: Node = head0
    node1: Node = head0.next
    node2: Node = head0.next.next
    node3: Node = head0.next.next.next
    node4: Node = head0.next.next.next.next

    # Execution
    removed0: Node = remove(target=node2)

    # Validation
    assert removed0 is node2
    assert node0.next is node1
    assert node0.previous is node4
    assert node1.next is node3
    assert node1.previous is node0
    assert node3.next is node4
    assert node3.previous is node1
    assert node4.next is node0
    assert node4.previous is node3
    assert node2.next is node2
    assert node2.previous is node2
    assert is_consistent(node=node0) is True
    assert is_consistent(node=node1) is True
    assert is_consistent(node=node2) is True
    assert is_consistent(node=node3) is True
    assert is_consistent(node=node4) is True
    assert is_consistent(node=removed0) is True



import pytest

from src.py_cdll import CDLL, NoAdjacentValueError, EmptyCDLLError


def test_before_unique_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)

    # Validation
    assert list0.before_unique(data2) is data1


def test_before_unique_single_entry_list_failure():
    # Setup
    data0: str = "data0"
    list0: CDLL = CDLL()
    list0.append(data0)

    # Validation
    with pytest.raises(NoAdjacentValueError):
        list0.before_unique(data0)


def test_before_unique_empty_list_failure():
    # Setup
    data0: str = "data0"
    list0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        list0.before_unique(data0)



from src.py_cdll import CDLL
from src.py_cdll.node import Node


def test__node_replace_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    new_node: Node = Node("new_data")
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._node_replace(target=cdll0._head.next, replacement=new_node)

    # Validation
    assert cdll0._head.next is new_node
    assert cdll0._head.next.next.value == data2
    assert cdll0._head.next.previous.value == data0


def test__node_replace_reassign_head_on_overwrite_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    new_node: Node = Node("new_data")
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0._node_replace(target=cdll0._head, replacement=new_node)

    # Validation
    assert cdll0._head is new_node
    assert cdll0._head.next.next.value == data2
    assert cdll0._head.previous.value == data2





from src.py_cdll import CDLL


def test___init___data_list_empty_success():
    # Setup
    datas0: list = []

    # Execution
    list0: CDLL = CDLL(values=datas0)

    # Validation
    assert len(list0) == 0


def test___init___data_list_one_item_success():
    # Setup
    data0: str = "data0"
    datas0: list = [data0]

    # Execution
    list0: CDLL = CDLL(values=datas0)

    # Validation
    assert len(list0) == 1
    assert list0[0] == data0


def test___init___data_list_four_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list = [data0, data1, data2, data3]

    # Execution
    list0: CDLL = CDLL(values=datas0)

    # Validation
    assert len(list0) == 4



from enum import Enum, auto
from typing import Sequence, Generic, TypeVar, MutableSequence, Iterable, overload, Callable, Any, Iterator

from .node import Node, merge_sort, reverse_order, insert_between, length, remove, nodes, update_head
from .compare import Comparison, compare
from .exceptions import ValueNotFoundError, MultipleValuesFoundError, \
    NoAdjacentValueError, NegativeIndexError, UnevenListLengthError, EmptyCDLLError, \
    FirstValueNotFoundError, SecondValueNotFoundError, ValuesNotAdjacentError, InputNotIterableError, \
    NoBeforeAndAfterUniqueError, CDLLAlreadyPopulatedError, AmbiguousEmptyCDLLStateError, \
    UnableToPopulateWithNoValuesError


class Connectivity(Enum):
    ADJACENT_NEXT = auto()
    ADJACENT_PREVIOUS = auto()


_T = TypeVar("_T")


class CDLL(MutableSequence[_T], Generic[_T]):

    def __init__(self, values: Sequence[_T] | None = None) -> None:
        self._head: Node | None = None
        self._length: int = 0

        if values is not None and len(values) > 0:
            self._populate(values=values)

    @property
    def _last(self) -> Node:
        if self.is_empty():
            raise EmptyCDLLError(f"Impossible to find last in empty CDLL.")

        return self._head.previous

    def __repr__(self) -> str:
        # values: list[str] = [repr(node.value) for _, node in self._nodes()]
        # string: str = f"CDLL[{len(self)}, [{', '.join(values)}]]"
        string: str = f"CDLL(...{id(self) % 1000})[{len(self)}, [{', '.join([])}]]"
        return string

    def __str__(self) -> str:
        values: list[str] = [repr(node.value) for _, node in self._nodes()]
        string: str = f"[{', '.join(values)}]"
        return string

    def is_empty(self) -> bool:
        is_length_zero: bool = len(self) == 0
        is_head_none: bool = self._head is None
        is_empty: bool = False

        if not is_head_none and not is_length_zero:
            is_empty = False
        elif is_head_none and is_length_zero:
            is_empty = True
        elif (is_head_none and not is_length_zero) or (not is_head_none and is_length_zero):
            raise AmbiguousEmptyCDLLStateError(f"Combination of head '{self._head}' and length '{len(self)}' values "
                                               f"makes it unclear whether CDLL is empty or not.")
        return is_empty

    def _add(self, first: Iterable[_T] | None, second: Iterable[_T] | None, is_inplace: bool) -> "CDLL[_T]":
        if not isinstance(first, Iterable):
            raise TypeError(f"Can only concatenate an Iterable (not '{type(first)}') to '{type(self)}'.")
        if not isinstance(second, Iterable):
            raise TypeError(f"Can only concatenate an Iterable (not '{type(second)}') to '{type(self)}'.")

        if is_inplace:
            cdll: CDLL[_T] = self
        else:
            cdll: CDLL[_T] = CDLL()

        if not is_inplace:
            try:
                cdll.extend(values=first)
            except EmptyCDLLError:
                pass  # When cdll is empty, there is nothing to add to the new cdll

        try:
            cdll.extend(values=second)
        except EmptyCDLLError:
            pass  # When cdll is empty, there is nothing to add to the new cdll

        return cdll

    def __add__(self, other: Iterable[_T]) -> "CDLL[_T]":
        new_list: CDLL[_T] = self._add(first=self, second=other, is_inplace=False)

        return new_list

    def __radd__(self, other: Iterable[_T]) -> "CDLL[_T]":
        new_list: CDLL[_T] = self._add(first=other, second=self, is_inplace=False)

        return new_list

    def __iadd__(self, other: Iterable[_T]) -> "CDLL[_T]":
        new_list: CDLL[_T] = self._add(first=self, second=other, is_inplace=True)

        return new_list

    def __mul__(self, other: int) -> "CDLL[_T]":
        if not isinstance(other, int):
            raise TypeError(f"List repetition is only possible with integers, not '{type(other)}'.")

        new_list: CDLL[_T] = CDLL()

        if other > 0:
            for _ in range(other):
                new_list.extend(values=self)

        return new_list

    def __rmul__(self, other: int) -> "CDLL[_T]":
        new_list: CDLL[_T] = self.__mul__(other=other)
        return new_list

    def __imul__(self, other: int) -> "CDLL[_T]":
        if not isinstance(other, int):
            raise TypeError(f"List repetition is only possible with integers, not '{type(other)}'.")

        if other > 1:
            values: CDLL[_T] = self[:len(self)]
            for _ in range(other - 1):
                self.extend(values=values)
        elif other == 1:
            pass
        elif other < 1:
            self.clear()

        return self

    @property
    def head(self) -> _T:
        if self.is_empty():
            raise EmptyCDLLError(f"Impossible to find head in empty CDLL.")

        return self._head.value

    @property
    def last(self) -> _T:
        if self.is_empty():
            raise EmptyCDLLError(f"Impossible to find last in empty CDLL.")

        return self._last.value

    def _node_remove(self, target: Node) -> None:
        if len(self) > 1:
            # Set new head before removing node to prevent head from temporarily pointing to a node
            # that is no longer in the correct list.
            if target is self._head:
                self._head = self._head.next

            _ = remove(target=target)
            self._length -= 1
        else:
            self.clear()

    def _node_remove_at_index(self, index: int) -> Node:
        target: Node = self._node_at_index(index=index)
        self._node_remove(target=target)
        return target

    # TODO: combine with _nodes_reverse() and add argument "reverse: bool = False" to select between them.
    def _nodes(self) -> Iterator[tuple[int, Node]]:
        if self.is_empty():
            return

        # TODO: refactor all methods relying on index to use node.index instead
        for index, node in enumerate(nodes(head=self._head, callback=lambda: self._head)):
            yield index, node

    def _nodes_reverse(self) -> Iterator[tuple[int, Node]]:
        if self.is_empty():
            return

        index: int = len(self) - 1
        current: Node = self._last

        while True:
            yield index, current

            index -= 1
            current = current.previous

            # Python list is strictly index based, and this shall reflect that
            # if current is self._head:
            if index < 0:
                break

    def _nodes_value(self,
                     value: _T,
                     comparison: Comparison = Comparison.IDENTITY_OR_EQUALITY
                     ) -> Iterator[tuple[int, Node]]:

        for index, node in self._nodes():
            is_comparable: bool = compare(pair=(value, node.value), comparison=comparison)

            if is_comparable:
                yield index, node

    def _nodes_index(self, indices: list[int], reverse: bool = False) -> Iterator[tuple[int, Node]]:
        indices.sort(reverse=reverse)

        if reverse:
            iterator: Iterator[tuple[int, Node]] = self._nodes_reverse()
        else:
            iterator: Iterator[tuple[int, Node]] = self._nodes()

        for index, node in iterator:
            while len(indices) > 0 and indices[0] == index:
                yield index, node
                indices = indices[1:]

    def _nodes_value_constrained(self,
                                 value: _T,
                                 start: int = 0,
                                 stop: int | None = None,
                                 comparison: Comparison = Comparison.IDENTITY_OR_EQUALITY
                                 ) -> Iterator[tuple[int, Node]]:

        if start < 0:
            start = max(len(self) + start, 0)

        if stop is None:
            stop = len(self)
        elif stop < 0:
            stop = max(len(self) + stop, 0)

        if start == stop:
            return

        for index, node in self._nodes():
            if start <= index <= stop:
                is_comparable: bool = compare(pair=(value, node.value), comparison=comparison)

                if is_comparable:
                    yield index, node

    def _node_unique(self, value: _T, comparison: Comparison = Comparison.IDENTITY) -> tuple[int, Node]:
        if self.is_empty():
            raise EmptyCDLLError(f"Impossible to find value in empty CDLL.")

        unique_amount: int = 0
        unique_index: int | None = None
        unique_node: Node | None = None

        for index, node in self._nodes_value(value=value, comparison=comparison):
            unique_amount += 1
            unique_index = index
            unique_node = node

        if unique_amount < 1:
            raise ValueNotFoundError(f"No instance of value found.")
        elif unique_amount > 1:
            raise MultipleValuesFoundError(f"More than one instance of value found.")

        return unique_index, unique_node

    def _node_first(self, value: _T, comparison: Comparison = Comparison.IDENTITY_OR_EQUALITY) -> tuple[int, Node]:
        if self.is_empty():
            raise EmptyCDLLError(f"Impossible to find value in empty CDLL.")

        try:
            index, node = next(self._nodes_value(value=value, comparison=comparison))
        except StopIteration as exception:
            raise ValueNotFoundError(f"No instance of value found") from exception

        return index, node

    def index_unique(self, value: _T) -> int:
        index, _ = self._node_unique(value=value, comparison=Comparison.IDENTITY)

        return index

    def index(self, value: _T, start: int = 0, stop: int | None = None) -> int:
        if start < 0:
            start = max(len(self) + start, 0)

        if stop is None:
            stop = len(self)
        elif stop < 0:
            stop = max(len(self) + stop, 0)

        try:
            first_index, _ = next(self._nodes_value_constrained(value=value,
                                                                start=start,
                                                                stop=stop,
                                                                comparison=Comparison.IDENTITY_OR_EQUALITY))
        except StopIteration as exception:
            raise ValueError(f"Value '{value}' not in list, so there is no index to return.") from exception

        return first_index

    def before_unique(self, value: _T) -> _T:
        if self.is_empty():
            raise EmptyCDLLError(f"Impossible to find value in empty CDLL.")
        elif len(self) < 2:
            raise NoAdjacentValueError(f"No unique adjacent value when length is less than 2.")

        _, node = self._node_unique(value=value, comparison=Comparison.IDENTITY)

        return node.previous.value

    def after_unique(self, value: _T) -> _T:
        if self.is_empty():
            raise EmptyCDLLError(f"Impossible to find value in empty CDLL.")
        elif len(self) < 2:
            raise NoAdjacentValueError(f"No unique adjacent value when length is less than 2.")

        _, node = self._node_unique(value=value, comparison=Comparison.IDENTITY)

        return node.next.value

    def before_and_after_unique(self, value: _T) -> tuple[_T, _T]:
        if self.is_empty():
            raise EmptyCDLLError(f"Impossible to find value in empty CDLL.")
        elif len(self) < 3:
            raise NoAdjacentValueError(f"No unique adjacent value when length is less than 3.")

        try:
            _, node = self._node_unique(value=value, comparison=Comparison.IDENTITY)
        except (ValueNotFoundError, MultipleValuesFoundError) as exception:
            raise NoBeforeAndAfterUniqueError(f"Could not get values around element in CDLL.") from exception

        before: _T = node.previous.value
        after: _T = node.next.value

        return before, after

    def opposite_unique(self, value: _T) -> _T:
        is_length_even: bool = len(self) % 2 == 0

        if not is_length_even:
            raise UnevenListLengthError(f"No opposite item in list of uneven length.")

        half_length: int = len(self) // 2
        value_index: int = self.index_unique(value=value)
        opposite_index: int = self._wraparound_index(index=value_index + half_length)
        opposite: _T = self[opposite_index]

        return opposite

    def replace_unique(self, value: _T, replacement: _T) -> None:
        _, node = self._node_unique(value=value, comparison=Comparison.IDENTITY)
        node.value = replacement

    def replace_before_unique(self, value: _T, replacement: _T) -> None:
        if len(self) == 1:
            raise NoAdjacentValueError(f"No value before input in list of length 1.")

        _, node = self._node_unique(value=value, comparison=Comparison.IDENTITY)
        node.previous.value = replacement

    def replace_after_unique(self, value: _T, replacement: _T) -> None:
        if len(self) == 1:
            raise NoAdjacentValueError(f"No value after input in list of length 1.")

        _, node = self._node_unique(value=value, comparison=Comparison.IDENTITY)
        node.next.value = replacement

    def remove(self, value: _T) -> None:
        try:
            _, node = self._node_first(value=value, comparison=Comparison.IDENTITY_OR_EQUALITY)
        except ValueNotFoundError as exception:
            raise ValueError(f"Value '{value}' not in list, so there is nothing to remove.") from exception

        self._node_remove(target=node)

    def remove_unique(self, value: _T) -> None:
        _, node = self._node_unique(value=value, comparison=Comparison.IDENTITY)
        self._node_remove(target=node)

    def pop(self, index: int | None = None) -> _T:
        if index is None:
            index = len(self) - 1

        try:
            target: Node = self._node_remove_at_index(index=index)
        except EmptyCDLLError as exception:
            raise IndexError(f"Can't pop a value from an empty cdll.") from exception

        value: _T = target.value

        return value

    def count(self, value: _T) -> int:
        iterator: Iterator[tuple[int, Node]] = self._nodes_value(value=value,
                                                                 comparison=Comparison.IDENTITY_OR_EQUALITY)
        amount: int = sum(1 for _ in iterator)
        return amount

    def sort(self, key: Callable[[_T], Any] | None = None, reverse: bool = False) -> None:
        if self.is_empty():
            # If head is None, there is nothing to sort.
            return

        head: Node = self._head

        if key is None:
            self._head = merge_sort(head=head)
        else:
            try:
                decorated: CDLL[tuple[Any, _T]] = CDLL(values=[(key(item), item) for item in self])
            except (KeyError, TypeError) as exception:
                raise exception

            decorated.sort()
            undecorated: CDLL[_T] = CDLL(values=[item for _, item in decorated])
            self._head = undecorated._head

        if reverse:
            self._head = reverse_order(head=self._head)

    def copy(self) -> "CDLL[_T]":
        duplicate: CDLL[_T] = CDLL()
        duplicate.extend(values=self)
        return duplicate

    # TODO: Evaluate for removal
    def adjacent(self, pair: tuple[_T, _T]) -> bool:
        # TODO: raise exception if length is <2

        is_nodes_adjacent: bool = False

        try:
            _, value0_node = self._node_unique(value=pair[0], comparison=Comparison.IDENTITY)
        except ValueNotFoundError:
            return is_nodes_adjacent

        is_value0_next_value1: bool = value0_node.next.value is pair[1]
        is_value0_previous_value1: bool = value0_node.previous.value is pair[1]
        is_nodes_adjacent = is_value0_next_value1 or is_value0_previous_value1

        return is_nodes_adjacent

    # TODO: Evaluate for removal
    def adjacency_direction(self, pair: tuple[_T, _T]) -> Connectivity:
        # Direction from first towards second, if adjacent
        connectivity: Connectivity | None = None

        try:
            _, first_node = self._node_unique(value=pair[0], comparison=Comparison.IDENTITY)
        except ValueNotFoundError as exception:
            raise FirstValueNotFoundError(f"First value not found in CDLL.") from exception

        if first_node.next.value is pair[1]:
            connectivity: Connectivity = Connectivity.ADJACENT_NEXT
        elif first_node.previous.value is pair[1]:
            connectivity: Connectivity = Connectivity.ADJACENT_PREVIOUS

        try:
            _, second_node = self._node_unique(value=pair[1], comparison=Comparison.IDENTITY)
        except ValueNotFoundError as exception:
            raise SecondValueNotFoundError(f"Second value not found in CDLL.") from exception

        if first_node and second_node and not connectivity:
            raise ValuesNotAdjacentError(f"'{pair[0]}' and '{pair[1]}' are not adjacent in '{self}'.")

        return connectivity

    def _populate(self, values: Sequence[_T]) -> None:
        if self.is_empty():
            values_iterator: Iterator[_T] = iter(values)

            try:
                self._head = Node(value=next(values_iterator))
            except StopIteration as exception:
                raise UnableToPopulateWithNoValuesError(f"Unable to populate CDLL "
                                                        f"from an empty Sequence") from exception

            self._length += 1

            self.extend(values=values_iterator)
        else:
            raise CDLLAlreadyPopulatedError(f"Unable to populate CDLL "
                                            f"as there is already '{len(self)}' elements in it.")

    def append(self, value: _T) -> None:
        if self.is_empty():
            self._populate(values=[value])
        else:
            insert: Node = Node(value=value)
            self._insert_after_node(anchor=self._last, insert=insert)

    def extend(self, values: Iterable[_T] | Sequence[_T]) -> None:
        try:
            for value in values:
                self.append(value=value)
        except TypeError as exception:
            raise InputNotIterableError(f"Input contains no value that can be appended to CDLL.") from exception

    def insert(self, index: int, value: _T) -> None:
        if self.is_empty():
            self._populate(values=[value])
        else:
            self._insert(index=index, value=value)

    def clear(self) -> None:
        if self.is_empty():
            return

        self._head = None
        self._length = 0

    # TODO: Evaluate for removal
    def _shift_head_forwards(self, amount: int = 1) -> None:
        # TODO: generalise shift head to take positive and negative
        # TODO: raise exception when cdll does not have even length
        #       can't seem to remember reason why that would be necessary in all cases...
        if self.is_empty():
            raise EmptyCDLLError(f"Not possible to shift head in empty CDLL.")

        if amount <= 0:
            raise ValueError(f"Amount to shift '{amount}' must be a positive integer")

        new_head: Node = self._head

        for _ in range(amount):
            new_head = new_head.next

        update_head(node=new_head)
        self._head = new_head

    # TODO: Evaluate for removal
    def _shift_head_backwards(self, amount: int = 1) -> None:
        # TODO: raise exception when cdll does not have even length
        #       can't seem to remember reason why that would be necessary in all cases...
        if self.is_empty():
            raise EmptyCDLLError(f"Not possible to shift head in empty CDLL.")

        if amount <= 0:
            raise ValueError(f"Amount to shift '{amount}' must be a positive integer")

        new_head: Node = self._head

        for _ in range(amount):
            new_head = new_head.previous

        update_head(node=new_head)
        self._head = new_head

    # TODO: Evaluate for removal
    def rotate(self, amount: int = 0) -> None:
        if amount == 0:
            pass
        elif amount > 0:
            self._shift_head_backwards(amount=amount)
        elif amount < 0:
            self._shift_head_forwards(amount=abs(amount))

    def _node_at_index(self, index: int) -> Node:
        if self.is_empty():
            raise EmptyCDLLError(f"Impossible to find value in empty CDLL.")

        index = self._normalize_index(index=index)

        for current_index, node in self._nodes():
            if current_index == index:
                current: Node = node
                break
        else:
            raise IndexError(f"Index '{index}' not in range of '{len(self)}'.")

        return current

    def _insert_after_node(self, anchor: Node, insert: Node) -> None:
        self._length += length(node=insert)
        insert_between(before=anchor, after=anchor.next, insert=insert)

    def _insert_before_node(self, anchor: Node, insert: Node) -> None:
        self._insert_after_node(anchor=anchor.previous, insert=insert)
        if anchor is self._head:
            self._head = insert

    def _node_replace(self, target: Node, replacement: Node) -> None:
        is_target_head: bool = target is target.head

        if len(self) > 1:
            head: Node | None = None
            if is_target_head:
                head = replacement

            _ = insert_between(before=target.previous, after=target.next, insert=replacement, head=head)

        if is_target_head:
            self._head = replacement

    def _replace_at_index(self, index: int, replacement: _T) -> None:
        # Imitates list behavior which raises an exception for any non-filled index.
        if self.is_empty():
            raise IndexError(f"Index value of '{index}' is not in range of size '({-len(self)}:)0:{len(self) - 1}'")
        else:
            target: Node = self._node_at_index(index=index)
            replacement: Node = Node(value=replacement)
            self._node_replace(target=target, replacement=replacement)

    def _replace_with_slice(self, segment: slice, values: Sequence[_T]) -> None:
        section: range = self._range_from_slice(segment=segment)
        indices: list[int] = list(section)
        are_indices_consecutive: bool = self._are_consecutive(numbers=indices)

        if are_indices_consecutive:
            for index in reversed(indices):
                self._node_remove_at_index(index=index)
            for value in reversed(values):
                try:
                    self._insert(index=section.start, value=value)
                except IndexError:
                    self.append(value=value)
        else:
            if len(indices) == len(values):
                for (_, node), value in zip(self._nodes_index(indices=indices), values):
                    replacement: Node = Node(value=value)
                    self._node_replace(target=node, replacement=replacement)
            else:
                raise ValueError(f"attempt to assign sequence of size '{len(values)}' "
                                 f"to extended slice of size '{len(indices)}'.")

    def _remove_with_slice(self, segment: slice) -> None:
        indices: list[int] = list(self._range_from_slice(segment=segment))
        indices.sort()

        for index in reversed(indices):
            self._node_remove_at_index(index=index)

    def _range_from_slice(self, segment: slice) -> range:
        start, stop, step = segment.indices(len(self))
        indices: range = range(start, stop, step)
        return indices

    @staticmethod
    def _are_consecutive(numbers: Sequence[int]) -> bool:
        are_consecutive: list[bool] = [numbers[index] + 1 == numbers[index + 1] for index in range(len(numbers) - 1)]
        are_all_consecutive: bool = all(are_consecutive)
        return are_all_consecutive

    def _insert(self, index: int, value: _T) -> None:
        # Imitates list behavior which inserts any index for an empty list at index zero,
        # and appends for any index out of range.
        if self.is_empty():
            self._populate(values=[value])
        else:
            try:
                node_current = self._node_at_index(index=index)
                node_new = Node(value=value)
                self._insert_before_node(anchor=node_current, insert=node_new)
            except IndexError:
                self.append(value=value)

    def __len__(self) -> int:
        return self._length

    def _normalize_index(self, index: int) -> int:
        if index < -len(self) or index >= len(self):
            # IndexError used by python loops to know when end of iterable has been reached
            raise IndexError(f"Index value of '{index}' is not in range of '({-len(self)}:)0:{len(self) - 1}'")

        if index < 0:
            index = index % len(self)

        return index

    def _wraparound_index(self, index: int) -> int:
        return index % len(self)

    def __eq__(self, other: "CDLL[_T]") -> bool:
        equal: bool = isinstance(other, type(self))

        if equal:
            equal &= len(self) == len(other)

        if equal:
            equal &= all(self_item is other_item or self_item == other_item
                         for self_item, other_item in zip(self, other))

        return equal

    # TODO: Evaluate for removal
    def _eq_rotated_mirrored(self, other: "CDLL[_T]") -> bool:
        # TODO: Do I really need the mirrored part of the check?
        # TODO: seems like this is broken when there are multiple items of the same in a list, and it is rotated
        #       then there is no guarantee that the find_first picks the correct of the two
        #       there would have to be a find_all method and a test against all the following items
        #       from each starting point until confirmed match, failure and go to next, or all failure

        self_length: int = len(self)

        if self_length == 0:
            return True

        # Shortcut for when both lists are uninitialized
        if self.is_empty() and other.is_empty():
            return True

        has_same_lengths: bool = self_length == len(other)
        has_same_items: bool = False

        if has_same_lengths:
            mirrored: bool = False
            connection_pair_equality: list[bool] = []

            start_item: _T = self.head
            other_start_index: int

            try:
                other_start_index = other.index(value=start_item)
            except ValueNotFoundError:
                return False

            # check start neighbors and their mirroring
            if self_length > 1:
                self_last: _T = self[-1]
                self_second: _T = self[1]

                other_last: _T = other[other._wraparound_index(index=other_start_index - 1)]
                other_second: _T = other[other._wraparound_index(index=other_start_index + 1)]

                all_none: bool = self_second is None and self_last is None and \
                                 other_second is None and other_last is None

                if self_second == other_last and self_last == other_second and not all_none:
                    mirrored = True

            connection_pairs: list[(_T, _T)] = []
            for index, element in enumerate(self):
                if mirrored:
                    index_other: int = other._wraparound_index(index=other_start_index - index)
                else:
                    index_other: int = other._wraparound_index(index=other_start_index + index)
                connection_pairs.append((self[index], other[index_other]))

            for element0, element1 in connection_pairs:
                connection_pair_equality.append(element0 == element1)

            # TODO: change comparisons to be added up with &=
            has_same_items: bool = all(connection_pair_equality)

        return has_same_lengths and \
               has_same_items

    # TODO: Evaluate for removal
    def mirror(self, index: int = 0) -> None:
        # TODO: evaluate if the normalize call can be removed
        index = self._normalize_index(index=index)

        if index < 0:
            raise NegativeIndexError(f"Index '{index}' is not zero or greater")

        if index > 0:
            self._shift_head_forwards(amount=index)

        pair_amount: int = (len(self) - 1) // 2
        # pairs are first and last remaining indexes after index 0, narrowing inwards,
        # until zero or one unassigned indexes remain
        index_pairs: list[tuple[int, int]] = [(1 + index, len(self) - 1 - index) for index in range(pair_amount)]

        for pair in index_pairs:
            self.switch(pair=pair)

        if index > 0:
            self._shift_head_backwards(amount=index)

    # TODO: Evaluate for removal
    def switch(self, pair: tuple[int, int]) -> None:
        self[pair[0]], self[pair[1]] = self[pair[1]], self[pair[0]]

    @overload
    def __getitem__(self, index: int) -> _T:
        ...

    @overload
    def __getitem__(self, index: slice) -> "CDLL[_T]":
        ...

    def __getitem__(self, index: int | slice) -> _T | "CDLL[_T]":
        if isinstance(index, int):
            value: _T = self._node_at_index(index=index).value
        elif isinstance(index, slice):
            value: CDLL[_T] = self._slice(segment=index)
        else:
            raise TypeError(f"__getitem__ requires an integer or a slice, not a {type(index)}.")
        return value

    @overload
    def __setitem__(self, index: int, value: _T) -> None:
        ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[_T]) -> None:
        ...

    def __setitem__(self, index: int | slice, value: _T | Iterable[_T]) -> None:
        if isinstance(index, int):
            self._replace_at_index(index=index, replacement=value)
        elif isinstance(index, slice):
            self._replace_with_slice(segment=index, values=value)
        else:
            raise TypeError(f"__setitem__ requires an integer or a slice, not a {type(index)}.")

    def __delitem__(self, index: int | slice) -> None:
        if isinstance(index, int):
            _: Node = self._node_remove_at_index(index=index)
        elif isinstance(index, slice):
            self._remove_with_slice(segment=index)
        else:
            raise TypeError(f"__delitem__ requires an integer or a slice, not a {type(index)}.")

    def __contains__(self, value: _T) -> bool:
        contains: bool = True

        try:
            _, _ = self._node_first(value=value, comparison=Comparison.IDENTITY_OR_EQUALITY)
        except (EmptyCDLLError, ValueNotFoundError):
            contains = False

        return contains

    def __reversed__(self) -> Iterator[_T]:
        for _, node in self._nodes_reverse():
            yield node.value

    def __iter__(self) -> Iterator[_T]:
        for _, node in self._nodes():
            yield node.value

    def _slice(self, segment: slice) -> "CDLL[_T]":
        cdll: CDLL[_T] = CDLL()

        indices: range = self._range_from_slice(segment=segment)
        reverse: bool = True if indices.step < 0 else False

        for _, node in self._nodes_index(indices=list(indices), reverse=reverse):
            cdll.append(value=node.value)

        return cdll


# TODO: Evaluate for removal
CDLLPair = tuple[CDLL, CDLL]
CDLLPairs = list[CDLLPair]


if __name__ == '__main__':
    pass



import pytest

from src.py_cdll import NotANodeError
from src.py_cdll.node import Node, length, insert_between, is_consistent


def test_length_not_node_failure():
    # Setup
    node0: float = 3.14

    # Validation
    with pytest.raises(NotANodeError):
        # noinspection PyTypeChecker
        length(node=node0)


def test_length_one_node_success():
    # Setup
    node0: Node = Node(value=None)
    length0: int = 1

    # Execution
    length1: int = length(node=node0)

    # Validation
    assert length0 == length1
    assert is_consistent(node=node0) is True


def test_length_two_nodes_success():
    # Setup
    node0: Node = Node(value=None)
    node1: Node = Node(value=None)
    insert_between(before=node0, after=node0, insert=node1)
    length0: int = 2

    # Execution
    length1: int = length(node=node0)

    # Validation
    assert length0 == length1
    assert is_consistent(node=node0) is True


def test_length_three_nodes_success():
    # Setup
    node0: Node = Node(value=None)
    node1: Node = Node(value=None)
    node2: Node = Node(value=None)
    insert_between(before=node0, after=node0, insert=node1)
    insert_between(before=node1, after=node0, insert=node2)
    length0: int = 3

    # Execution
    length1: int = length(node=node0)

    # Validation
    assert length0 == length1
    assert is_consistent(node=node0) is True


def test_length_four_nodes_success():
    # Setup
    node0: Node = Node(value=None)
    node1: Node = Node(value=None)
    node2: Node = Node(value=None)
    node3: Node = Node(value=None)
    insert_between(before=node1, after=node1, insert=node2)
    insert_between(before=node0, after=node0, insert=node1)
    insert_between(before=node2, after=node0, insert=node3)
    length0: int = 4

    # Execution
    length1: int = length(node=node0)

    # Validation
    assert length0 == length1
    assert is_consistent(node=node0) is True



from src.py_cdll import CDLL


def test_count_in_empty_failure():
    # Setup
    data: int = 1
    cdll0: CDLL = CDLL()
    count0: int = 0

    # Execution
    count1: int = cdll0.count(value=data)

    # Validation
    assert count0 == count1


def test_count_no_instance_among_single_success():
    # Setup
    data: int = 1
    data0: int = 0
    datas0: list[int] = [data0]
    cdll0: CDLL = CDLL(values=datas0)
    count0: int = 0

    # Execution
    count1: int = cdll0.count(value=data)

    # Validation
    assert count0 == count1


def test_count_one_instance_among_single_success():
    # Setup
    data: int = 1
    data0: int = 1
    datas0: list[int] = [data0]
    cdll0: CDLL = CDLL(values=datas0)
    count0: int = 1

    # Execution
    count1: int = cdll0.count(value=data)

    # Validation
    assert count0 == count1


def test_count_no_instance_among_multiple_success():
    # Setup
    data: int = 11
    data0: int = 0
    data1: int = 1
    data2: int = 2
    data3: int = 3
    data4: int = 4
    data5: int = 5
    datas0: list[int] = [data0, data1, data2, data3, data4, data5]
    cdll0: CDLL = CDLL(values=datas0)
    count0: int = 0

    # Execution
    count1: int = cdll0.count(value=data)

    # Validation
    assert count0 == count1


def test_count_one_instance_among_multiple_success():
    # Setup
    data: int = 1
    data0: int = 0
    data1: int = 1
    data2: int = 2
    data3: int = 3
    data4: int = 4
    data5: int = 5
    datas0: list[int] = [data0, data1, data2, data3, data4, data5]
    cdll0: CDLL = CDLL(values=datas0)
    count0: int = 1

    # Execution
    count1: int = cdll0.count(value=data)

    # Validation
    assert count0 == count1


def test_count_four_instances_among_multiple_success():
    # Setup
    data: int = 1
    data0: int = 0
    data1: int = 1
    data2: int = 1
    data3: int = 3
    data4: int = 1
    data5: int = 1
    datas0: list[int] = [data0, data1, data2, data3, data4, data5]
    cdll0: CDLL = CDLL(values=datas0)
    count0: int = 4

    # Execution
    count1: int = cdll0.count(value=data)

    # Validation
    assert count0 == count1



import pytest

from src.py_cdll import CDLL


def test___iadd___wrong_type_failure():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)
    adder0: int = 22

    # Validation
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        cdll0 += adder0


def test___iadd___list_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas0 + datas0)
    cdll2: CDLL = cdll0

    # Execution
    cdll0 += datas0

    # Validation
    assert cdll1 == cdll0
    assert cdll2 is cdll0


def test___iadd___empty_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas0)
    cdll2: CDLL = cdll0

    # Execution
    cdll0 += []

    # Validation
    assert cdll1 == cdll0
    assert cdll2 is cdll0


def test___iadd___one_value_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += cdll1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___two_values_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)
    cdll3: CDLL = cdll0

    # Execution
    # cdll3: CDLL = (
    cdll0 += cdll1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___five_values_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    data7: str = "data7"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3, data4, data5, data6, data7]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)
    cdll3: CDLL = cdll0

    # Execution
    # cdll3: CDLL =
    cdll0 += cdll1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___item_and_empty_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = [data0]
    datas1: list[str] = []
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0)
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += cdll1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___empty_and_item_success():
    # Setup
    data0: str = "data0"
    datas0: list[str] = []
    datas1: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas1)
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += cdll1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___item_and_item_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    datas0: list[str] = [data0]
    datas1: list[str] = [data1]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += cdll1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___two_items_and_five_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1]
    datas1: list[str] = [data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += cdll1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___three_items_and_four_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2]
    datas1: list[str] = [data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas1)
    cdll2: CDLL = CDLL(values=datas0 + datas1)
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += cdll1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___cdll_tuple_success():
    # Setup
    cdll0: CDLL = CDLL()
    tuple1: tuple = tuple()
    cdll2: CDLL = CDLL()
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += tuple1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___cdll_list_success():
    # Setup
    cdll0: CDLL = CDLL()
    list1: list = list()
    cdll2: CDLL = CDLL()
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += list1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___cdll_set_success():
    # Setup
    cdll0: CDLL = CDLL()
    set1: set = set()
    cdll2: CDLL = CDLL()
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += set1

    # Validation
    assert cdll2 == cdll0
    assert cdll3 is cdll0


def test___iadd___cdll_range_success():
    # Setup
    cdll0: CDLL = CDLL()
    cdll1: CDLL = CDLL(values=[0, 1, 2, 3, 4])
    range1: range = range(5)
    cdll3: CDLL = cdll0

    # Execution
    cdll0 += range1

    # Validation
    assert cdll1 == cdll0
    assert cdll3 is cdll0



import pytest

from src.py_cdll import CDLL


def test___rmul___zero_empty_success():
    # Setup
    multiplier0: int = 0

    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]

    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL()

    # Execution
    cdll2: CDLL = multiplier0 * cdll0

    # Validation
    assert cdll1 == cdll2
    assert cdll2 is not cdll0


def test___rmul___one_same_success():
    # Setup
    multiplier0: int = 1

    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]

    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas0)

    # Execution
    cdll2: CDLL = multiplier0 * cdll0

    # Validation
    assert cdll1 == cdll2
    assert cdll2 is not cdll0


def test___rmul___two_double_success():
    # Setup
    multiplier0: int = 2

    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]

    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas0 + datas0)

    # Execution
    cdll2: CDLL = multiplier0 * cdll0

    # Validation
    assert cdll1 == cdll2
    assert cdll2 is not cdll0


def test___rmul___five_quintuple_success():
    # Setup
    multiplier0: int = 5

    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]

    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL(values=datas0 + datas0 + datas0 + datas0 + datas0)

    # Execution
    cdll2: CDLL = multiplier0 * cdll0

    # Validation
    assert cdll1 == cdll2
    assert cdll2 is not cdll0


def test___rmul___minus_one_empty_success():
    # Setup
    multiplier0: int = -1

    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]

    cdll0: CDLL = CDLL(values=datas0)
    cdll1: CDLL = CDLL()

    # Execution
    cdll2: CDLL = multiplier0 * cdll0

    # Validation
    assert cdll1 == cdll2
    assert cdll2 is not cdll0


def test___rmul___non_integer_failure():
    # Setup
    multiplier0: float = 3.7

    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]

    cdll0: CDLL = CDLL(datas0)

    # Validation
    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        multiplier0 * cdll0



from src.py_cdll.node import Node, nodes_from_values, length_step, is_consistent


def test_length_one_from_head_step_success():
    # Setup
    values0: list[str] = ["zero"]
    node0: Node = nodes_from_values(values=values0)
    amount0: int = 1

    # Execution
    amount1: int = length_step(node=node0)

    # Validation
    assert amount0 == amount1
    assert is_consistent(node=node0) is True


def test_length_two_from_head_step_success():
    # Setup
    values0: list[str] = ["zero", "one"]
    node0: Node = nodes_from_values(values=values0)
    amount0: int = 2

    # Execution
    amount1: int = length_step(node=node0)

    # Validation
    assert amount0 == amount1
    assert is_consistent(node=node0) is True


def test_length_three_from_head_step_success():
    # Setup
    values0: list[str] = ["zero", "one", "two"]
    node0: Node = nodes_from_values(values=values0)
    amount0: int = 3

    # Execution
    amount1: int = length_step(node=node0)

    # Validation
    assert amount0 == amount1
    assert is_consistent(node=node0) is True


def test_length_four_from_last_step_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = node0.previous
    amount0: int = 4

    # Execution
    amount1: int = length_step(node=node1)

    # Validation
    assert amount0 == amount1
    assert is_consistent(node=node0) is True


def test_length_five_from_middle_step_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = node0.next.next
    amount0: int = 5

    # Execution
    amount1: int = length_step(node=node1)

    # Validation
    assert amount0 == amount1
    assert is_consistent(node=node0) is True



from src.py_cdll.node import Node, stitch, is_consistent, nodes_from_values


def test_stitch_single_node_list_success():
    # Setup
    node0: Node = Node(value=None)
    node0.next = None
    node0.previous = None

    # Verification
    assert node0.next is None
    assert node0.previous is None

    # Execution
    stitch(head=node0, last=node0)

    # Validation
    assert node0.next is node0
    assert node0.previous is node0
    assert is_consistent(node=node0) is True


def test_stitch_double_node_list_success():
    # Setup
    node0: Node = nodes_from_values(values=["zero", "one"])
    node1: Node =  node0.next
    node0.previous = None
    node1.next = None

    # Verification
    assert node0.next is node1
    assert node0.previous is None
    assert node1.next is None
    assert node1.previous is node0

    # Execution
    stitch(head=node0, last=node1)

    # Validation
    assert node0.next is node1
    assert node0.previous is node1
    assert node1.next is node0
    assert node1.previous is node0
    assert is_consistent(node=node0) is True
    assert is_consistent(node=node1) is True



import pytest

from src.py_cdll import CDLL


def test___delitem___index_positive_in_range_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    index0: int = 2

    # Execution
    del cdll0[index0]

    # Validation
    assert data2 not in cdll0
    assert cdll0._head.value is data0
    assert cdll0._head.next.value is data1
    assert cdll0._head.next.next.value is data3
    assert cdll0._head.next.next.next.value is data4
    assert cdll0._head.next.next.next is cdll0._last
    assert cdll0._last.previous.previous.previous is cdll0._head


def test___delitem___index_negative_in_range_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    index0: int = -2

    # Execution
    del cdll0[index0]

    # Validation
    assert data3 not in cdll0
    assert cdll0._head.value is data0
    assert cdll0._head.next.value is data1
    assert cdll0._head.next.next.value is data2
    assert cdll0._head.next.next.next.value is data4
    assert cdll0._head.next.next.next is cdll0._last
    assert cdll0._last.previous.previous.previous is cdll0._head


def test___delitem___index_positive_out_of_range_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    index0: int = 11

    # Validation
    with pytest.raises(IndexError):
        del cdll0[index0]


def test___delitem___index_negative_out_of_range_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)
    index0: int = -8

    # Validation
    with pytest.raises(IndexError):
        del cdll0[index0]



from typing import Callable

import pytest

from src.py_cdll import CDLL


def test_sort_integers_success():
    # Setup
    list0: list[int] = [71, 52, 28, 33, 35, 25, 70, 36, 27, 6]
    list1: list[int] = [6, 25, 27, 28, 33, 35, 36, 52, 70, 71]
    cdll0: CDLL[int] = CDLL(values=list0)
    cdll1: CDLL[int] = CDLL(values=list1)

    # Execution
    cdll0.sort()

    # Validation
    assert cdll0 == cdll1


def test_sort_strings_success():
    # Setup
    # noinspection SpellCheckingInspection
    list0: list[str] = ['glpwi', 'atmev', 'rlcar', 'ayupu', 'nezfv', 'xdeur', 'tpdht', 'lrcjt', 'odswj', 'cdzyo']
    # noinspection SpellCheckingInspection
    list1: list[str] = ['atmev', 'ayupu', 'cdzyo', 'glpwi', 'lrcjt', 'nezfv', 'odswj', 'rlcar', 'tpdht', 'xdeur']
    cdll0: CDLL[str] = CDLL(values=list0)
    cdll1: CDLL[str] = CDLL(values=list1)

    # Execution
    cdll0.sort()

    # Validation
    assert cdll0 == cdll1


def test_sort_functions_failure():
    # Setup
    list0: list[Callable] = [max, min, all, sorted, reversed, sum, any]
    cdll0: CDLL[Callable] = CDLL(values=list0)

    # Validation
    with pytest.raises(TypeError):
        cdll0.sort()


def test_sort_dict_key_correct_success():
    # Setup
    list0: list[dict[str, int]] = [{'age': 16, 'cats': 3}, {'age': 33, 'cats': 10}, {'age': 37, 'cats': 5},
                                   {'age': 43, 'cats': 6}, {'age': 26, 'cats': 10}, {'age': 80, 'cats': 0},
                                   {'age': 30, 'cats': 9}, {'age': 65, 'cats': 10}, {'age': 86, 'cats': 1}]
    list1: list[dict[str, int]] = [{'age': 16, 'cats': 3}, {'age': 26, 'cats': 10}, {'age': 30, 'cats': 9},
                                   {'age': 33, 'cats': 10}, {'age': 37, 'cats': 5}, {'age': 43, 'cats': 6},
                                   {'age': 65, 'cats': 10}, {'age': 80, 'cats': 0}, {'age': 86, 'cats': 1}]
    cdll0: CDLL[dict[str, int]] = CDLL(values=list0)
    cdll1: CDLL[dict[str, int]] = CDLL(values=list1)

    # Execution
    cdll0.sort(key=lambda x: x["age"])

    # Validation
    assert cdll0 == cdll1


def test_sort_dict_key_incorrect_failure():
    # Setup
    list0: list[dict[str, int]] = [{'age': 16, 'cats': 3}, {'age': 33, 'cats': 10}, {'age': 37, 'cats': 5},
                                   {'age': 43, 'cats': 6}, {'age': 26, 'cats': 10}, {'age': 80, 'cats': 0},
                                   {'age': 30, 'cats': 9}, {'age': 65, 'cats': 10}, {'age': 86, 'cats': 1}]
    cdll0: CDLL[dict[str, int]] = CDLL(values=list0)

    # Validation
    with pytest.raises(KeyError):
        cdll0.sort(key=lambda x: x["height"])


def test_sort_dict_key_not_subscriptable_failure():
    # Setup
    list0: list[Callable] = [max, min, all, sorted, reversed, sum, any]
    cdll0: CDLL[Callable] = CDLL(values=list0)

    # Validation
    with pytest.raises(TypeError):
        cdll0.sort(key=lambda x: x["some_key"])


def test_sort_integers_reversed_success():
    # Setup
    list0: list[int] = [71, 52, 28, 33, 35, 25, 70, 36, 27, 6]
    list1: list[int] = [71, 70, 52, 36, 35, 33, 28, 27, 25, 6]
    cdll0: CDLL[int] = CDLL(values=list0)
    cdll1: CDLL[int] = CDLL(values=list1)

    # Execution
    cdll0.sort(reverse=True)

    # Validation
    assert cdll0 == cdll1



import pytest

from src.py_cdll import CDLL, EmptyCDLLError


def test_last_empty_list_failure():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        _ = cdll0.last


def test_last_success():
    # Setup
    data0: str = "head"
    data1: str = "end"
    datas0: list[str] = [data0, data1]
    cdll0: CDLL = CDLL(values=datas0)

    # Validation
    assert cdll0.head == data0
    assert cdll0.last == data1



from src.py_cdll import CDLL


def test___contains___empty_cdll_success():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL()

    # Execution
    bool0: bool = data0 in cdll0

    # Validation
    assert bool0 is False


def test___contains___data_not_in_cdll_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=[data0])

    # Execution
    bool0: bool = data1 in cdll0

    # Validation
    assert bool0 is False


def test___contains___data_in_cdll_success():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL(values=[data0])

    # Execution
    bool0: bool = data0 in cdll0

    # Validation
    assert bool0 is True



import pytest

from src.py_cdll import CDLL, NoAdjacentValueError


def test_replace_before_unique_overwrite_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)

    # Execution
    list0.replace_before_unique(value=data0, replacement=data)

    # Validation
    assert list0.before_unique(data0) is data


def test_replace_before_unique_single_entry_list_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    list0: CDLL = CDLL()
    list0.append(data0)

    # Validation
    with pytest.raises(NoAdjacentValueError):
        list0.replace_before_unique(value=data0, replacement=data1)



class BaseCDLLException(Exception):
    """Base exception for Circular Doubly Linked List project."""


class CDLLAlreadyPopulatedError(BaseCDLLException):
    """Raise when attempting to populate CDLL, but there is already data in it."""


class NegativeIndexError(BaseCDLLException):
    """Raise for indices that should be at least 0."""


class ValueNotFoundError(BaseCDLLException):
    """Raise for searches that return no result."""


class FirstValueNotFoundError(BaseCDLLException):
    """Raise when the first input of a method was not found where expected."""


class SecondValueNotFoundError(BaseCDLLException):
    """Raise when the second input of a method was not found where expected."""


class ValuesNotAdjacentError(BaseCDLLException):
    """Raise when values were not found adjacent to each other."""


class NoAdjacentValueError(BaseCDLLException):
    """Raise for searches for adjacent values in lists with single item."""


class MultipleValuesFoundError(BaseCDLLException):
    """Raise for searches that return no result."""


class UnableToPopulateWithNoValuesError(BaseCDLLException):
    """Raise when attempting to populate CDLL without any values."""


class EmptyCDLLError(BaseCDLLException):
    """Raise when an iterable is found to be empty."""


class UnevenListLengthError(BaseCDLLException):
    """Raise for lists that contain uneven amount of items."""


class InputNotIterableError(BaseCDLLException):
    """Raise when an input is not iterable."""


class NoBeforeAndAfterUniqueError(BaseCDLLException):
    """Raise when it is impossible to return the values before and after a unique element."""


class NoOrderedOccurrenceError(BaseCDLLException):
    """Raise when it is impossible to tell if data is ordered because values are missing."""


class AmbiguousEmptyCDLLStateError(BaseCDLLException):
    """Raise when it is ambiguous whether a CDLL is empty."""


class NotANodeError(BaseCDLLException):
    """Raise when a supposed Node turns out not to be so."""


class NodesNotInSameListError(BaseCDLLException):
    """Raise when two Nodes are not in the same list."""


class NoNewHeadError(BaseCDLLException):
    """Raise when no new head is provided for Node sequence."""



import pytest

from src.py_cdll import CDLL, InputNotIterableError


def test_extend_input_empty_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list = []
    datas1: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas1)
    cdll1: CDLL = CDLL(values=datas1)

    # Execution
    cdll0.extend(values=datas0)

    # Validation
    assert cdll0 == cdll1


def test_extend_input_one_item_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data3]
    datas1: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=datas1)
    cdll1: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll0.extend(values=datas0)

    # Validation
    assert cdll0 == cdll1


def test_extend_input_six_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data6]
    datas1: list[str] = [data0, data1, data2, data3, data4, data5]
    cdll0: CDLL = CDLL(values=datas1)
    cdll1: CDLL = CDLL(values=datas1 + datas0)

    # Execution
    cdll0.extend(values=datas0)

    # Validation
    assert cdll0 == cdll1


def test_extend_input_non_valid_failure():
    # Setup
    data0: int = 9
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    datas1: list[str] = [data0, data1, data2, data3, data4, data5]
    cdll0: CDLL = CDLL(values=datas1)

    # Validation
    with pytest.raises(InputNotIterableError):
        # noinspection PyTypeChecker
        cdll0.extend(values=data0)



import pytest

from src.py_cdll import CDLL, EmptyCDLLError, MultipleValuesFoundError, ValueNotFoundError


def test_remove_unique_empty_cdll_failure():
    # Setup
    data: str = "data"
    cdll0: CDLL = CDLL()

    # Execution
    with pytest.raises(EmptyCDLLError):
        cdll0.remove_unique(value=data)


def test_remove_unique_single_instance_among_single_success():
    # Setup
    data: str = "data"
    data0: str = "data"
    datas0: list[str] = [data0]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0.remove_unique(value=data)

    # Verification
    assert len(cdll0) == 0


def test_remove_unique_single_instance_among_multiple_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    cdll0.remove_unique(value=data)

    # Verification
    assert data not in cdll0


def test_remove_unique_multiple_instances_among_multiple_failure():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data"
    data2: str = "data2"
    data3: str = "data"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Verification
    with pytest.raises(MultipleValuesFoundError):
        cdll0.remove_unique(value=data)


def test_remove_unique_no_instances_among_multiple_failure():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    datas0: list[str] = [data0, data1, data2, data3, data4]
    cdll0: CDLL = CDLL(values=datas0)

    # Verification
    with pytest.raises(ValueNotFoundError):
        cdll0.remove_unique(value=data)



import pytest

from src.py_cdll import CDLL


def test___setitem___index_looping_over_and_replacing_all_success():
    # Setup
    data0: int = 37
    data1: int = 73
    cdll0: CDLL = CDLL(values=[data0, data1])

    # Execution
    for index, _ in enumerate(cdll0):
        cdll0[index] = index

    # Validation
    assert cdll0._head.value == 0
    assert cdll0._last.value == 1
    assert cdll0._head.next is cdll0._last
    assert cdll0._head.previous is cdll0._last


def test___setitem___index_in_range_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=[data0])

    # Execution
    cdll0[0] = data1

    # Validation
    assert cdll0._head.value is data1
    assert cdll0._head.next is cdll0._head
    assert cdll0._head.previous is cdll0._head


def test___setitem___index_out_of_range_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=[data0])

    # Validation
    with pytest.raises(IndexError):
        cdll0[1] = data1


def test___setitem___index_empty_list_out_of_range_failure():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(IndexError):
        cdll0[0] = data0


def test___setitem___index_negative_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1]
    cdll0: CDLL = CDLL(values=datas0)

    # Verification
    assert cdll0._head.value is data0
    assert cdll0._last.value is data1
    assert cdll0._head.next is cdll0._last
    assert cdll0._head.previous is cdll0._last
    assert cdll0._last.next is cdll0._head
    assert cdll0._last.previous is cdll0._head

    # Execution
    cdll0[-1] = data2

    # Validation
    assert cdll0._head.value is data0
    assert cdll0._last.value is data2
    assert cdll0._head.next is cdll0._last
    assert cdll0._head.previous is cdll0._last
    assert cdll0._last.next is cdll0._head
    assert cdll0._last.previous is cdll0._head



from src.py_cdll import CDLL


def test___str___empty_success():
    # Setup
    cdll0: CDLL = CDLL()

    # Validation
    assert cdll0.__str__() == "[]"


def test___str___text_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL[str] = CDLL(values=datas0)
    repr0: str = f"['{data0}', '{data1}', '{data2}']"

    # Validation
    assert cdll0.__str__() == repr0


def test___str___int_success():
    # Setup
    data0: int = 0
    data1: int = 1
    data2: int = 2
    datas0: list[int] = [data0, data1, data2]
    cdll0: CDLL[int] = CDLL(values=datas0)
    repr0: str = f"[{data0}, {data1}, {data2}]"

    # Validation
    assert cdll0.__str__() == repr0


def test___str___float_success():
    # Setup
    data0: float = 0.1
    data1: float = 1.2
    data2: float = 2.3
    datas0: list[float] = [data0, data1, data2]
    cdll0: CDLL[float] = CDLL(values=datas0)
    repr0: str = f"[{data0}, {data1}, {data2}]"

    # Validation
    assert cdll0.__str__() == repr0



import pytest

from src.py_cdll import CDLL, EmptyCDLLError
from src.py_cdll.exceptions import NoBeforeAndAfterUniqueError, NoAdjacentValueError


def test_before_and_after_unique_three_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)

    # Execution
    previous0, next0 = list0.before_and_after_unique(value=data1)

    # Validation
    assert previous0 == data0 and next0 == data2


def test_before_and_after_unique_six_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)
    list0.append(data3)
    list0.append(data4)
    list0.append(data5)

    # Execution
    previous0, next0 = list0.before_and_after_unique(value=data4)

    # Validation
    assert previous0 == data3 and next0 == data5


def test_before_and_after_unique_two_items_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)

    # Validation
    with pytest.raises(NoAdjacentValueError):
        list0.before_and_after_unique(value=data1)


def test_before_and_after_unique_one_item_success():
    # Setup
    data0: str = "data0"
    list0: CDLL = CDLL()
    list0.append(data0)

    # Validation
    with pytest.raises(NoAdjacentValueError):
        list0.before_and_after_unique(value=data0)


def test_before_and_after_unique_zero_items_failure():
    # Setup
    data0: str = "data0"
    list0: CDLL = CDLL()

    # Validation
    with pytest.raises(EmptyCDLLError):
        list0.before_and_after_unique(value=data0)


def test_before_and_after_unique_four_items_not_found_failure():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)
    list0.append(data3)

    # Validation
    with pytest.raises(NoBeforeAndAfterUniqueError):
        list0.before_and_after_unique(value=data)


def test_before_and_after_unique_seven_items_multiple_found_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    list0: CDLL = CDLL()
    list0.append(data0)
    list0.append(data1)
    list0.append(data2)
    list0.append(data3)
    list0.append(data4)
    list0.append(data3)
    list0.append(data3)

    # Validation
    with pytest.raises(NoBeforeAndAfterUniqueError):
        list0.before_and_after_unique(value=data3)



from src.py_cdll.node import Node, nodes_from_values, equal, update_head, is_consistent


def test_update_head_one_item_first_success():
    # Setup
    values0: list[str] = ["zero"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = nodes_from_values(values=values0)
    node2: Node = node0.next

    # Execution
    update_head(node=node2)

    # Validation
    assert equal(first=node2, second=node1)
    assert is_consistent(node=node2) is True


def test_update_head_two_item_second_success():
    # Setup
    values0: list[str] = ["zero", "one"]
    values1: list[str] = ["one", "zero"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = nodes_from_values(values=values1)
    node2: Node = node0.next

    # Execution
    update_head(node=node2)

    # Validation
    assert equal(first=node2, second=node1)
    assert is_consistent(node=node2) is True


def test_update_head_three_item_second_success():
    # Setup
    values0: list[str] = ["zero", "one", "two"]
    values1: list[str] = ["one", "two", "zero"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = nodes_from_values(values=values1)
    node2: Node = node0.next

    # Execution
    update_head(node=node2)

    # Validation
    assert equal(first=node2, second=node1)
    assert is_consistent(node=node2) is True


def test_update_head_four_item_third_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three"]
    values1: list[str] = ["two", "three", "zero", "one"]
    node0: Node = nodes_from_values(values=values0)
    node1: Node = nodes_from_values(values=values1)
    node2: Node = node0.next.next

    # Execution
    update_head(node=node2)

    # Validation
    assert equal(first=node2, second=node1)
    assert is_consistent(node=node2) is True



from typing import Iterator

from src.py_cdll import CDLL
from src.py_cdll.node import Node


def test__nodes_index_empty_nodes_empty_indices_success():
    # Setup
    indices0: list[int] = []
    cdll0: CDLL = CDLL()

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_index(indices=indices0)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_index_empty_nodes_five_indices_success():
    # Setup
    indices0: list[int] = [2, 3, 6, 9, 14]
    cdll0: CDLL = CDLL()

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_index(indices=indices0)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_index_four_nodes_empty_indices_success():
    # Setup
    indices0: list[int] = []
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    datas0: list[str] = [data0, data1, data2, data3]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_index(indices=indices0)

    # Validation
    assert len(list(iterator0)) == 0


def test__nodes_index_seven_nodes_three_indices_success():
    # Setup
    indices0: list[int] = [2, 3, 6]
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_index(indices=indices0)

    # Validation
    assert len(list(iterator0)) == 3


def test__nodes_index_seven_nodes_five_repeated_indices_success():
    # Setup
    indices0: list[int] = [2, 3, 3, 6, 6]
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_index(indices=indices0)

    # Validation
    assert len(list(iterator0)) == 5


def test__nodes_index_seven_nodes_three_indices_also_out_of_range_success():
    # Setup
    indices0: list[int] = [2, 11, 37]
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    data4: str = "data4"
    data5: str = "data5"
    data6: str = "data6"
    datas0: list[str] = [data0, data1, data2, data3, data4, data5, data6]
    cdll0: CDLL = CDLL(values=datas0)

    # Execution
    iterator0: Iterator[tuple[int, Node]] = cdll0._nodes_index(indices=indices0)

    # Validation
    assert len(list(iterator0)) == 1
